/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Code.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import java.io.*;
import uiuc.Jumbo.Jaemus.*;

/**
 * An interface for all pieces that make up a classfile that may have holes.
 * @version      1.0,  2 Oct 2001
 * @author       Lars Clausen
 *
*/

public abstract class Code {
    public ClassEnvironment defineClasses(ClassEnvironment env) {
        return env;
    }
    public Environment defineSupers(Environment env) {
        return env;
    }
    public Environment defineMembers(Environment env) {
        return env;
    }
    public abstract ClosedCode eval(Environment env);
    public ClosedCode evalLHS(final Code code, final int op, Environment env) {
        throw new Error("Trying to call evalLHS on "+this);
    }
    public ClosedCode evalLHSNoUse(final Code code, final int op,
                                      Environment env) {
        throw new Error("Trying to call evalLHSNoUse on "+this);
    }
    /** Evaluate the code, but assume that nobody will be using the result. */
    public ClosedCode evalNoUse(Environment env) {
        ClosedCode c = eval(env);
        if (c.type != Type.void_type && c.type != Type.bottom)
            return new ClosedCode(c.bytecode.addPop(c.type),
                                  env, Type.void_type);
        else
            return c;
    }
    /** Evaluate the code, and perform appropriate jumps.
     *  Note that if the code is constant, the result will have
     *  a boolean constant value, even though its type is void.
     */
    public ClosedCode evalWithLabels(Environment env,
                                     Instr truelabel,
                                     Instr falselabel,
                                     Instr nextlabel) {
        ClosedCode c = eval(env);
        if (c.type == Type.boolean_type) {
            if (c.isConstant()) {
                Instr targetlabel =
                    (c.constantValue().booleanVal()?truelabel:falselabel);
                if (nextlabel != targetlabel)
                    return new ClosedCode(c.bytecode.
                                          addGoto(targetlabel),
                                          env, Type.void_type,
                                          c.constantValue());
                else return new ClosedCode(c.bytecode, env, Type.void_type,
                                           c.constantValue());
            }
            // The default case has truelabel == nextlabel.  If not so, fix it.
            Instr jumplabel = falselabel;
            int cmp = Instr.CMP_EQ;
            Instr gotoinstr = Instr.genNil("Fall through");
            if (truelabel != nextlabel) {
                jumplabel = truelabel;
                cmp = Instr.CMP_NE;
                // If we need a goto, it will be to falselabel, because
                // truelabel uses the jump
                if (falselabel != nextlabel)
                    gotoinstr = Instr.genGoto(falselabel);
            }
            return new ClosedCode(c.bytecode.
                                  addIf(cmp, jumplabel).
                                  append(gotoinstr),
                                  env, Type.void_type);
        } else {
            throw new Error("evalWithLabels of non-boolean value "+this);
        }
    }

    /** Evaluate this code as RHS of a plus expression.
     * @param env The current environment.
     */
    public ClosedCode evalInAdd(Environment env) {
        return eval(env);
    }

    /** Generate any classes in this piece of code. 
     * @deprecated
     */
    public void load() {
        generate();
    }
    /** Generate any classes in this piece of code. */
    public void generate() {
        ClassEnvironment e = defineClasses(new ClassEnvironment());
        eval(defineMembers(defineSupers(new Environment(e))));
    }

    /** Generate code and load the class. */
    public Object create(String classname) {
        generate();
        try {
            Class c = Class.forName(classname);
            return c.newInstance();
        } catch (Exception e) {
            throw new Error(e.toString());
        }
    }
    /** Answers whether this piece of code is a call to a constructor on
        the same class.  Necessary to figure out whether to include the
        instance variable initializers in a constructor */
    boolean isThisConstructorCall() {
        return false;
    }
}

/* End of interface Code*/
