/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Environment.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.util.*;
import java.io.*;
import uiuc.Jumbo.Util.*;

/**
 * An environment for constructors, providing variables and labels.  This file has to move.
 * @version      1.0, 18 Oct 2001
 * @author       Lars Clausen
 *
*/

public class Environment extends CodeEnvironment implements Cloneable {
    private final Local locals;
    private final int var_index;
    private final ControlLabel labels;
    
    public Environment() {
        locals = null;
        var_index = 0;
        labels = null;
    }

    public Environment(ClassEnvironment e) {
        super(e);
        locals = null;
        var_index = 0;
        labels = null;
    }

    public Environment(Environment e, ClassInfo c, Local free) {
        super(e, c, free);
        locals = null;
        var_index = 0;
        labels = null;
    }

    /* Environment creator for method declaration */
    private Environment(Environment old, MethodInfo method, MonoList args) {
        super(old, method);
        labels = old.labels;

        Local newlocals = null;
        int newvar_index = 0;
        if (!method.flagsSet(Flags.STATIC)) {
            newlocals = new Var("this", newvar_index++, 
                                new Type(getCurrentClass().getName()),
                                Flags.FINAL, old.locals);
        } else {
            newlocals = old.locals;
        }
        if (args != null)
            for (Iterator i = args.iterator(); i.hasNext(); ) {
                ArgDef arg = (ArgDef)i.next();
                Local param = new Var(arg.name,
                                      newvar_index++,
                                      Constructor.normalizeClassNames(arg.type, this), 
                                      arg.flags, newlocals);
                if (param.isLargeType()) newvar_index++;
                newlocals = param;
            }
        locals = newlocals;
        var_index = newvar_index;
    }

    public Environment startClass(ClassInfo c, Local free) {
        return new Environment(this, c, free);
    }

    public Environment startMethod(MethodInfo method, MonoList args) {
        return new Environment(this, method, args);
    }

    /* Used by defineVar */
    private Environment(Environment old, String name, Type type, int flags,
                        ConstVal cv) {
        super(old);
        labels = old.labels;

        locals = new Var(name, old.var_index, type, flags, old.locals, cv);
        if (locals.isLargeType()) var_index = old.var_index+2;
        else var_index = old.var_index+1;
    }

    private Environment(Environment old, String name, Type type, int flags) {
        super(old);
        labels = old.labels;
        var_index = old.var_index;

        locals = new Local(name, type, flags, old.locals);
    }

    public Environment defineVar(String name, Type type) {
        return defineVar(name, type, 0, null);
    }

    public Environment defineVar(String name, Type type, int flags, 
                                 ConstVal cv) {
        return new Environment(this, name, type, flags, cv);
    }

    public Environment defineLocal(String name, Type type, int flags) {
        return new Environment(this, name, type, flags);
    }

    public Environment generateVar(String name, Type type) {
        return defineVar(name+"$"+var_index, type, 0, null);
    }

    public Environment generateVar(String name, Type type, int flags) {
        return defineVar(name+"$"+var_index, type, flags, null);
    }

    public Var lookupVar(String name) {
        name = name.intern();
        for (Local v = locals; v != null; v = v.next) {
            if (v.name == name && v instanceof Var) return (Var)v;
        }
        return null;
    }

    public Local lookupLocal(String name) {
        name = name.intern();
        for (Local v = locals; v != null; v = v.next) {
            if (v.name == name && !(v instanceof Var)) return v;
        }
        return null;
    }

    /** Return a newly defined variable.
     * This should only be used right after generateVar.
     */
    public Var getNewestVar() {
        return (Var)locals;
    }

    public boolean varDefined(String name) {
        return lookupVar(name) != null;
    }

    private Environment(Environment old, ControlLabel l) {
        super(old);
        labels = l;
        locals = old.locals;
        var_index = old.var_index;
    }

    public Environment defineBreakLabel(String label, Instr breakpoint) {
        ControlLabel l = new ControlLabel(ControlLabel.BREAK, label, breakpoint, labels);
        return new Environment(this, l);
    }
    public Environment defineContinueLabel(String label, Instr breakpoint) {
        ControlLabel l = new ControlLabel(ControlLabel.CONTINUE, label, breakpoint, labels);
        return new Environment(this, l);
    }
    public Environment defineFinallyLabel(Instr handler) {
        ControlLabel l = new ControlLabel(ControlLabel.FINALLY, "", handler, labels);
        return new Environment(this, l);
    }

    public MonoList getBreakLabel(String label) 
        throws LabelNotFoundException {
        MonoList ls = new ArrayMonoList();
        for (ControlLabel l = labels; l != null; l = l.next) {
            if (l.isFinallyLabel())
                ls.add(l.place);
            if (l.isBreakLabel(label)) {
                ls.add(l.place);
                return ls;
            }
        }
        throw new LabelNotFoundException(label);
    }

    public MonoList getContinueLabel(String label)
        throws LabelNotFoundException {
        MonoList ls = new ArrayMonoList();
        for (ControlLabel l = labels; l != null; l = l.next) {
            if (l.isFinallyLabel())
                ls.add(l.place);
            if (l.isContinueLabel(label)) {
                ls.add(l.place);
                return ls;
            }
        }
        throw new LabelNotFoundException(label);
    }

    public MonoList getFinallyLabels() {
        MonoList ls = new ArrayMonoList();
        for (ControlLabel l = labels; l != null; l = l.next) {
            if (l.isFinallyLabel())
                ls.add(l.place);
        }
        return ls;
    }

    public ClassInfo get(String classname) throws ClassNotFoundException {
        ClassInfo c = getCurrentClass();
        if (c != null) {
            if (c.hasInnerClasses() &&
                getCurrentMethod() != null) {
                Local l = lookupLocal(classname);
                if (l != null) {
                    return getClass(l.type.getClassName());
                }
            }
            ClassInfo ci = c.getInnerClass(classname);
            if (ci != null) return ci;
            if (selfIsInner()) {
                return ((Environment)getOuter()).get(classname);
            }
        }
        return super.get(classname);
    }

    public Environment copy() {
        try {
            Environment clone = (Environment)super.clone();
            return clone;
        } catch (CloneNotSupportedException e) {
            return null;
        }
    }
    public static Environment emptyEnv() {
        return new Environment();
    }
    public static Environment emptyOuterEnv() {
        Environment e = new Environment();
        return e;
    }
}

/* End of class Environment*/

class ControlLabel {
    static final int BREAK = 0;
    static final int CONTINUE = 1;
    static final int FINALLY = 2;
    final int type;
    final String name;
    final Instr place;
    final ControlLabel next;

    public ControlLabel(int type, String name, Instr place, ControlLabel next) {
        this.type = type; 
        this.name = name;
        this.place = place;
        this.next = next;
    }
    public boolean isBreakLabel(String label) {
        if (type != BREAK) return false;
        return (label.equals("") || name.equals(label));
    }
    public boolean isContinueLabel(String label) {
        if (type != CONTINUE) return false;
        return (label.equals("") || name.equals(label));
    }
    public boolean isFinallyLabel() {
        return type == FINALLY;
    }
    public String toString() {
        switch (type) {
        case BREAK: return "break "+name;
        case CONTINUE: return "continue "+name;
        case FINALLY: return "finally";
        default: return "Unknown label type "+type;
        }
    }
}

class LabelNotFoundException extends Exception {
    final String label;
    public LabelNotFoundException(String label) {
        this.label = label;
    }
    public String toString() {
        if (label.equals(""))
            return "No label found";
        return "Label `"+label+"' not found";
    }
}
