/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Test.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.io.*;
import java.lang.reflect.*;
import uiuc.Jumbo.Util.*;

/**
 * Test framework for the constructors and related structures
 * Extends Constructor to avoid long names.
 * @version      1.0, 18 Oct 2001
 * @author       Lars Clausen
 *
*/

public class Test {
    static int total_tests = 0;
    static int failures = 0;
    static String test_name;
    static void assertEq(Object o1, Object o2) {
        if (!o1.equals(o2)) {
            System.out.println("Failure doing "+test_name+": "+o1+" != "+o2);
            failures++;
        }
    }

    static void unitTest0() {
        ClosedCode c = Constructor.constant(1).eval(Environment.emptyEnv());
        assertEq(c.bytecode, Instr.genConst(1));
        assertEq(c.type, Type.int_type);
    }

    static void unitTest1() {
        ClassInfo cl = new ClassInfo("Tests/Test1", "");
        MethodInfo mi = new MethodInfo(cl, 0, "dummy", 
                                       new MethodType(Type.void_type));
        ClosedCode c = Constructor.negate(Constructor.constant((long)1)).eval(Environment.emptyEnv());
        assertEq(c.bytecode, Instr.genConst((long)-1));
        assertEq(c.type, Type.long_type);
    }

    static void unitTest2() {
        makeTest(2, Constructor.constant(1), 1);
    }

    static void unitTest3() {
        makeTest(3, Constructor.mathOp(Instr.ADD, Constructor.constant(1), Constructor.constant(2)), 3);
    }

    static void unitTest4() {
        makeTest(4, Constructor.bitwiseOp(Instr.XOR, Constructor.constant(3), Constructor.constant(6)), 5);
    }

    static void unitTest5() {
        makeTest(5, Constructor.trinaryConditional(Constructor.booleanNot(Constructor.constant(true)),
                                       Constructor.constant(3), Constructor.constant(5)), 5);
    }

    static void unitTest6() {
        makeTest(6, Constructor.trinaryConditional(Constructor.booleanNot(Constructor.constant(false)),
                                       Constructor.constant(4), Constructor.constant(6)), 4);
    }

    static void unitTest7() {
        makeVarTest(7, "x", Constructor.putValue(Constructor.getQualifiedName(new ArrayMonoList("x")), Constructor.constant(3)), 3);
    }

    static void unitTest8() {
        makeTest(8, Constructor.arrayLength(Constructor.newArray(Type.int_type, new ArrayMonoList(Constructor.constant(10)))), 10);
    }

    private static void testFindMethod(String inclass, String name, MethodType type) {
        try {
            Environment.emptyEnv().findMethod(inclass, name, type);
        } catch (MethodNotFoundException e) {
            System.out.println("Failed to find method "+inclass+"."+name);
            failures++;
        }
    }

    static void unitTest9() {
        testFindMethod("java/lang/String", "hashCode", new MethodType(Type.void_type));
    }

    static void unitTest10() {
        testFindMethod("java/lang/Class", "forName", new MethodType(Type.object_type, Type.string_type));
    }

    static void unitTest11() {
        // Test that subclassing works.
        testFindMethod("java/lang/Class", "isInstance", new MethodType(Type.boolean_type, Type.string_type));
    }

    static void unitTest12() {
        testFindMethod("java/lang/reflect/Method", "invoke", new MethodType(Type.boolean_type, Type.object_type, new Type(Type.string_type)));
        
    }

    static void unitTest13() {
        // Test that we can find methods in superclasses
        testFindMethod("java/lang/String", "getClass", new MethodType(Type.boolean_type));        
    }

    static void unitTest14() {
        // Test that we can find a field
        if (Environment.emptyEnv().findField("java/lang/System", "err")
            == null)
            failures++;
    }

    static void unitTest15() {
        // Test that we can find a field in a superclass
        if (Environment.emptyEnv().findField("java/io/BufferedInputStream", "in")
            == null)
            failures++;
    }

    static void unitTest16() {
        // Test that we can't find a non-existant field
        if (Environment.emptyEnv().findField("java/lang/Object", "forname")
            != null)
            failures++;
    }

    /*
    static void unitTest17() {
        makeTest(17, getQualifiedName(new ArrayMonoList("java").add("lang").add("Integer").add("MAX_VALUE")), Integer.MAX_VALUE);
    }
    */

    private static void makeTest(int number, Code code, int result) {
        MethodInfo m = makeTestMethod("Tests/Test"+number, "testMethod");
        Environment env = Environment.emptyEnv();
        env.setCurrentMethod(m);
        ClosedCode cc = Constructor.returnVal(code).eval(env);
        m.addCode(cc.bytecode);
        testMethod(m, result);
    }

    private static void makeVarTest(int number, String name, Code code,
                                    int result) {
        MethodInfo m = makeTestMethod("Tests/Test"+number, "testMethod");
        Environment env = Environment.emptyEnv();
        env = env.startMethod(m, new ArrayMonoList());
        ClosedCode cc = Constructor.statements(new ArrayMonoList(Constructor.defineVariable(name, Type.int_type)).add(code).add(Constructor.returnVal(Constructor.getQualifiedName(new ArrayMonoList(name))))).
            eval(env);
        m.addCode(cc.bytecode);
        testMethod(m, result);
    }

    private static MethodInfo makeTestMethod(String classname, String methodname) {
        ClassInfo cl = new ClassInfo(classname, "");
        cl.flags = Flags.PUBLIC;
        MethodInfo m = new MethodInfo(cl, Flags.PUBLIC|Flags.STATIC, methodname, 
                                      new MethodType(Type.int_type));
        cl.addMethod(m);
        return m;
    }

    private static void testMethod(MethodInfo mi, int result) {
        ClassInfo cl = mi.getClassInfo();
        try {
            cl.writeClassFile();
            Class c = Class.forName(cl.getJavaName());
            if (c == null) {
                System.out.println("Failed to load "+cl.getJavaName());
                failures++;
                return;
            }
            Method m = c.getDeclaredMethod(mi.getName(), new Class[0]);
            if (m == null) {
                System.out.println("Method "+mi.getName()+" not in class "+cl.getJavaName());
                failures++;
                return;
            }
            int i = ((Integer)m.invoke(null, new Object[0])).intValue();
            if (i != result) {
                System.out.println("Method "+cl.getJavaName()+"."+mi.getName()+"() didn't return "+result+": "+i);
                failures++;
                return;
            }
            new File(cl.name+".class").delete();
        } catch (Exception e) {
            System.out.println("Exception creating "+cl.getJavaName()+": "+e);
            e.printStackTrace();
            failures++;
        }

    }

    static void unitTests() {
        int i = 0;
        try {
            Class c = Class.forName("uiuc.Jumbo.Compiler.Test");
            Class[] args = new Class[0];
            while (true) {
                test_name = "unitTest"+i;
                Method m = c.getDeclaredMethod(test_name, args);
                total_tests++;
                try {
                    m.invoke(null, args);
                } catch (InvocationTargetException e) {
                    System.out.println("Exception while executing "+m+": "+e);
                    e.printStackTrace();
                    failures++;
                }               
                i++;
            }
        } catch (NoSuchMethodException e) {
        } catch (ClassNotFoundException e) {
            System.out.println("Failed to find self class: "+e);
        } catch (IllegalAccessException e) {
            System.out.println("Can't get to method `"+test_name+"': "+e);
        } 
    }

    public static void main(String[] argv) {
        unitTests();
        if (failures > 0) {
            System.out.println("Failed "+failures+" of "+total_tests+" tests.");
        } else {
            System.out.println("All "+total_tests+" tests passed.");
        }
    }
}

/* End of class Test*/
