/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ClassInfo.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import uiuc.Jumbo.Util.*;

/**
 * A class for outputting the overall structure of a .class file
 * @version      1.0,  2 Oct 2001
 * @author       Lars Clausen
 *
*/

public class ClassInfo implements ClassMember {
    public String name;
    public String supername;
    public int flags;
    public ConstantPool constantpool = new ConstantPool();
    public MonoList implementees = new ArrayMonoList();
    public MonoList fields = new ArrayMonoList();
    public MonoList methods = new ArrayMonoList();
    public MonoList attributes = new ArrayMonoList();
    private MonoList innerclasses;
    public ClassInfo outerclass;
    private MonoList freevariables;

    int anon_id=1;
    Constant constant;
    
    String sourcefile;

    /** The directory to put class files in.  By default the current dir. */
    public static String outputdir = "";
    
    /** Construct a new class with the given name.
     *  The class has java.lang.Object as a superclass and
     *  default protection. */
    public ClassInfo(String name) {
        this.flags = Flags.SUPER;
        this.name = name;
        this.supername = "java/lang/Object";
    }

    public ClassInfo(int flags, String name) {
        this.name = name;
        this.supername = "java/lang/Object";
        this.flags = flags | Flags.SUPER;
    }

    /** Construct a new class with the given name and optional superclass name.
     *  Both names are in internal bytecode format, i.e. package parts are
     *  separated by '/' rather than '.'.  This also allows us to use the
     *  names directly for file names.
     *  If the supername is the empty string, the default superclass 
     *  (java/lang/Object) is used.  If it is null, the class has no superclass
     *  (is either java/lang/Object itself or an interface).
     */
    public ClassInfo(String name, String supername) {
        if ((supername == null || supername.equals("")) &&
            (!name.equals("java/lang/Object")))
            supername = "java/lang/Object";
        this.flags = Flags.SUPER;
        this.name = name;
        this.supername = supername;
    }
    public ClassInfo(int flags, String name, String supername) {
        this(name, supername);
        this.flags = flags | Flags.SUPER;
    }

    /** Construct a new class from a Java-style class object.
     */
    public ClassInfo(Class c) {
        this.name = c.getName().replace('.', '/');
        if (c.getSuperclass() != null)
            this.supername = c.getSuperclass().getName().replace('.', '/');
        else
            this.supername = null;
        Class[] interfaces = c.getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            this.implementsInterface(interfaces[i].getName().replace('.', '/'));
        }
        // Now transform methods, fields etc into our style.
        Method[] methods = null;
        try {
            methods = c.getDeclaredMethods();
        } catch (SecurityException e) {
            throw new Error("Can't access methods for "+name+": "+e);
        }
        for (int i = 0; i < methods.length; i++) {
            addMethod(new MethodInfo(this, methods[i]));
        }
        Constructor[] constructors = null;
        try {
            constructors = c.getDeclaredConstructors();
        } catch (SecurityException e) {
            throw new Error("Can't access constructors for "+name+": "+e);
        }
        for (int i = 0; i < constructors.length; i++) {
            addMethod(new MethodInfo(this, constructors[i]));
        }
        Field[] fields = null;
        try {
            fields = c.getDeclaredFields();
        } catch (SecurityException e) {
            throw new Error("Can't access fields for "+name+": "+e);
        }
        for (int i = 0; i < fields.length; i++) {
            addField(new FieldInfo(this, fields[i]));
        }
    }
    
    /** Constructs a new class with an optional outer class and flags.
     */
    public ClassInfo(ClassInfo outerclass, int flags, 
                     String name, String supername) {
        if ((supername == null || supername.equals("")) &&
            (!name.equals("java/lang/Object")))
        if (outerclass != null) { // Inner class
            outerclass.addInnerClass(this);
            if (name == null) { // Anonymous inner class
                name = outerclass.getName()+"$"+outerclass.getAnonId();
            } else {
                name = outerclass.getName()+"$"+name;
            }
            this.outerclass = outerclass;
        }
        this.name = name;
        this.flags = flags | Flags.SUPER;
        this.supername = supername;
    }

    /** Construct a new class from a class file stream.
     */
    public static ClassInfo readClass(InputStream in) 
        throws Exception {
        DataInputStream di = new DataInputStream(in);
        int count;

        int magic = di.readInt();
        if (magic != (int) 0xCAFEBABE) {
            throw new Error("Doesn't start with magic number 0xCAFEBABE:\n"+magic);
        }

        int minorVersion = di.readUnsignedShort();
        int majorVersion = di.readUnsignedShort();
        if (majorVersion < 45) {
            throw new Error("Strange version: "+majorVersion+"."+minorVersion);
        }

        /* Read constant pool */
        ConstantPool constantpool = new ConstantPool(di);
        int flags = di.readUnsignedShort();

        /* Read class name index */
        int thisindex = di.readUnsignedShort();
        String name = ((ConstantClass)constantpool.get(thisindex)).getName();

        int superindex = di.readUnsignedShort();
        String supername = null;
        if (superindex != 0) {
            supername = ((ConstantClass)constantpool.get(superindex)).getName();
        }

        ClassInfo c;
        if ((flags & Flags.INTERFACE) != 0) {
            c = new InterfaceInfo(flags, name);
        } else {
            c = new ClassInfo(flags, name, supername);
        }
        c.constantpool = constantpool;

        /* Identify all of the interfaces implemented by this class */
        count = di.readUnsignedShort();
        for (int i = 0; i < count; i++) {
            int iindex = di.readUnsignedShort();
            c.implementsInterface((((ConstantClass)c.constantpool.get(iindex)).getName()));
        }

        /*
         * Identify all fields in this class.
         */
        count = di.readUnsignedShort();
        c.fields = new ArrayMonoList(count);
        for (int i = 0; i < count; i++) {
            FieldInfo f = new FieldInfo(di, c.constantpool);
            f.inClass = c;
            c.addField(f);
        }

        /*
         * Identify all the methods in this class.
         */
        count = di.readUnsignedShort();
        c.methods = new ArrayMonoList(count);
        for (int i = 0; i < count; i++) {
            MethodInfo m = new MethodInfo(di, c.constantpool);
            m.inClass = c;
            c.addMethod(m);
        }

        /** We don't need the attributes of the class, so skip 'em */
        /*
         * Identify all of the attributes in this class
         */
        /*
          count = di.readUnsignedShort();
          attributes = new AttributeInfo[count];
          for (int i = 0; i < count; i++) {
          try {
          Message.push("Class attribute "+i);
          attributes[i] = AttributeInfo.readAttribute(di, constantPool, this);
          } catch (Exception anException) {
          Message.error("While reading class attribute "+i+":"+anException);
          return false;
          } finally {
          Message.pop();
          }
          }
        */

        return c;
    }

    public ClassInfo getSuperClass() {
        if (supername != null)
            return ClassTable.get(supername);
        else return null;
    }

    public void setSuperClass(String supername) {
        if (supername != null && supername.equals(""))
            supername = "java/lang/Object";
        this.supername = supername;
    }
    /** Return the full name of this class in Java style.
     *  All "/"'s are replaced with "."'s
     */
    public String getJavaName() {
        return name.replace('/', '.');
    }

    /** Return the full name of this class in bytecode style.
     */
    public String getName() {
        return name;
    }

    public int getAnonId() {
        return anon_id++;
    }

    public boolean flagSet(int flag) {
        return (flags & flag) == flag;
    }

    /** Adds a method to the class.
     * @param method MethodInfo representing the method to be added
     */
    public void addMethod(MethodInfo method) {
        methods.add(method);
    }

    /** Add a field to this class */
    public void addField(FieldInfo f) {
        // Could check for duplicates.
        fields.add(f);
    }

    public void implementsInterface(String i) {
        implementees.add(i);
    }

    public void setOuterClass(ClassInfo i) {
        outerclass = i;
    }
    
    public void addInterface(InterfaceInfo i) {
        if (innerclasses == null) innerclasses = new ArrayMonoList();
        innerclasses.add(i);
    }

    public void addInnerClass(ClassInfo i) {
        if (innerclasses == null) innerclasses = new ArrayMonoList();
        innerclasses.add(i);
    }

    public void addMember(MethodInfo method) {
        addMethod(method);
    }

    public void addMember(FieldInfo field) {
        addField(field);
    }

    public void addMember(ClassInfo inner) {
        addInnerClass(inner);
    }

    public void addMember(InterfaceInfo inner) {
        addInnerClass(inner);
    }

    public void addMember(ClassMember e) {
        e.getAdded(this);
    }

    /**
     * Get the value of sourcefile.
     * @return Value of sourcefile.
     */
    public String getSourcefile() {return sourcefile;}
    
    /**
     * Set the value of sourcefile.
     * @param v  Value to assign to sourcefile.
     */
    public void setSourcefile(String  v) {this.sourcefile = v;}
    
    /**
     * Get the value of constant.
     * @return Value of constant.
     */
    public Constant getConstant() {return constant;}
    
    /**
     * Set the value of constant.
     * @param v  Value to assign to constant.
     */
    public void setConstant(Constant  v) {this.constant = v;}

    public boolean hasInnerClasses() { return innerclasses != null; }

    public MonoList getInnerClasses() { return innerclasses; }

    public void addFreeVariable(String name) {
        if (freevariables == null) freevariables = new ArrayMonoList();
        freevariables.add(name);
    }

    public MonoList getFreeVariables() {
        return freevariables;
    }

    public boolean hasField(String name) {
        for (Iterator i = fields.iterator(); i.hasNext(); ) {
            FieldInfo f = (FieldInfo)i.next();
            if (f.name.equals(name)) return true;
        }
        return false;
    }

    /**
     * Returns the first inner class found with the given name.
     * @param n  Class name to look for, wittout $ mangling
     * @return  ClassInfo instance representing an inner class
     * of this class or a superclass.
     */
    public ClassInfo getInnerClass(String n) {
        if (innerclasses != null) {
            String dollarname = "$"+n;
            for (Iterator i = innerclasses.iterator(); i.hasNext(); ) {
                ClassInfo ci = (ClassInfo)i.next();
                if (ci.name.endsWith(dollarname)) 
                    return ci;
            }
        }
        if (supername != null) {
            return ClassTable.get(supername).getInnerClass(n);
        }
        return null;
    }

    public boolean hasConstructorMethod() {
        for (Iterator i = methods.iterator(); i.hasNext(); ) {
            MethodInfo m = (MethodInfo)i.next();
            if (m.getName().equals("<init>")) return true;
        }
        return false;
    }

    public boolean isInterface() { return false; }        

    /* Static classes nested in classes are not considered inner classes */
    public boolean isInnerClass() {
        return outerclass != null && (flags & Flags.STATIC) == 0;
    }

    public boolean isStatic() {
        return (flags & Flags.STATIC) != 0;
    }

    /** Turns the list of implementees into a list of InterfaceInfos.
     *  Initially, to allow for late definitions, they are stored as
     *  strings.  
     */
    protected void reifyImplementees() {
        if (implementees.size() == 0) return;
        if (implementees.get(0) instanceof InterfaceInfo) return;
        MonoList newimplementees = new ArrayMonoList();
        for (Iterator i = implementees.iterator(); i.hasNext(); ) {
            String in = (String)i.next();
            InterfaceInfo iface = new InterfaceInfo(in);
            newimplementees.add(iface);
        }
        implementees = newimplementees;
    }

    /** Returns true if this class is a subclass of the given class.
     */
    public boolean subclassOf(ClassInfo c) {
        if (equals(c)) return true;
        if (implementees != null) {
            for (Iterator i = implementees.iterator(); i.hasNext(); ) {
                String iface = (String)i.next();
                InterfaceInfo ic = (InterfaceInfo)ClassTable.get(iface);
                if (ic.subclassOf(c)) return true;
            }
        }
        if (isInterface()) {
            // An interface is always a subclass of Object.
            return c.getName().equals("java/lang/Object");
        }
        if (supername == null) return false;
        ClassInfo superclass = null;
        superclass = ClassTable.get(supername);
        if (superclass == null)
            return false;
        return superclass.subclassOf(c);
    }

    public void writeClassFile() throws IOException {
        if (!outputdir.equals("")) {
            File dir = new File(outputdir);
            dir.mkdirs();
        }
        File outputfile = new File(outputdir+name+".class");
        DataOutputStream o = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(outputfile)));
        prepareForEmit(constantpool);
        emitTo(o);
        o.close();
    }

    public byte[] writeArray() throws IOException {
        ByteArrayOutputStream b = new ByteArrayOutputStream();
        DataOutputStream o = new DataOutputStream(b);
        prepareForEmit(constantpool);
        emitTo(o);
        o.flush();
        byte[] bytes = b.toByteArray();
        o.close();
        return bytes;
    }

    public void prepareForEmit(ConstantPool c) {
        c.prepareForEmit();
        c.addClass(name);
        c.addClass(supername);

        attributes.add(new SourcefileAttribute(name+".java"));

        for (Iterator i = implementees.iterator(); i.hasNext();) {
            String iface = (String)i.next();
            c.addClass(iface);
        }

        for (Iterator i = fields.iterator(); i.hasNext();) {
            FieldInfo f = (FieldInfo)i.next();
            f.prepareForEmit(c);
        }

        for (Iterator i = methods.iterator(); i.hasNext();) {
            MethodInfo m = (MethodInfo)i.next();
            m.prepareForEmit(c);
        }

        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.prepareForEmit(c);
        }
    }

    public void emitTo(DataOutputStream o) throws IOException {
        o.writeInt(0xCAFEBABE);
        o.writeShort(3); // Minor version
        o.writeShort(45); // Major version

        constantpool.emitTo(o);

        o.writeShort(flags);
        constantpool.getClass(name).emitIndex(o);
        constantpool.getClass(supername).emitIndex(o);

        o.writeShort(implementees.size());
        for (Iterator i = implementees.iterator(); i.hasNext();) {
            String f = (String)i.next();
            constantpool.getClass(f).emitIndex(o);
        }

        o.writeShort(fields.size());
        for (Iterator i = fields.iterator(); i.hasNext();) {
            FieldInfo f = (FieldInfo)i.next();
            f.emitTo(o);
        }

        o.writeShort(methods.size());
        for (Iterator i = methods.iterator(); i.hasNext();) {
            MethodInfo f = (MethodInfo)i.next();
            f.emitTo(o);
        }

        o.writeShort(attributes.size());
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.emitTo(o);
        }

        /*
        if (innerclasses != null) {
            for (Iterator i = innerclasses.iterator(); i.hasNext();) {
                ClassInfo c = (ClassInfo)i.next();
                c.writeClassFile();
            }
        }
        */
    }

    public void getAdded(ClassInfo c) {
        c.addMember(this);
    }

    public boolean equals(Object o) {
        if (!(o instanceof ClassInfo)) return false;
        ClassInfo c = (ClassInfo)o;
        return name.equals(c.name);
    }

    public String toString() {
        return name;
    }
}

/* End of class ClassInfo*/
