/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Constant.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.io.*;

/**
 * A class describing a constant in a constant pool
 * @version      1.0,  5 Oct 2001
 * @author       Lars Clausen
 *
*/

public abstract class Constant {
    int index = -1;
    public Constant() {
    }

    static final int CLASS = 7;
    static final int FIELDREF = 9;
    static final int METHODREF = 10;
    static final int STRING = 8;
    static final int INTEGER = 3;
    static final int FLOAT = 4;
    static final int LONG = 5;
    static final int DOUBLE = 6;
    static final int INTERFACE = 11;
    static final int NAMEANDTYPE = 12;
    static final int UTF8 = 1;

    // Return true if this constant may be used by LDC, and thus should
    // be stored early in the constant pool
    public boolean isLdcConstant() {
        return false;
    }
    void resolve(Constant[] constants) {
    }
    public void onInsertion(ConstantPool c) {
    }
    void emitConstant(DataOutputStream o) 
        throws IOException {
        o.writeByte(constantNumber());
    }
    public void emitIndex(DataOutputStream o) throws IOException {
        if (index == -1) {
            throw new Error("Required constant "+this+" has no index");
        }
        o.writeShort(index);
    }
    public boolean isDoubleConstant() { return false; }
    public void emitByteIndex(DataOutputStream o) throws IOException {
        if (index == -1) {
            throw new Error("Required constant "+this+" has no index");
        }
        if (index > 255) {
            throw new Error("Required constant "+this+" at "+index+" has index > 255");
        }
        o.writeByte(index);
    }
    abstract public int constantNumber();
}
class ConstantClass extends Constant {
    String name;
    Constant name_constant;
    int indexbyte;
    ConstantClass(String name) {
        this.name = name;
    }
    ConstantClass(ClassInfo cl) {
        this.name = cl.getName();
    }
    ConstantClass(Type t) {
        this(t.className());
    }
    ConstantClass(DataInputStream di) 
        throws Exception {
        indexbyte = di.readUnsignedShort();
    }
    void resolve(Constant[] constants) {
        name_constant = constants[indexbyte];
        name_constant.resolve(constants);
        name = name_constant.toString();
    }
    public boolean isLdcConstant() { return true; }
    public void onInsertion(ConstantPool c) {
        name_constant = c.addUTF8(name);
    }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeShort(name_constant.index);
    }
    int getIndex() {
        return index;
    }

    public int hashCode() {
        return name.hashCode()^0x0ae21b47; // Random integer
    }
    public String toString() {
        return "Class "+name;
    }
    public String getName() {
        return name;
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantClass)) return false;
        ConstantClass cc = (ConstantClass)other;
        return name.equals(cc.name);
    }
    public int constantNumber() { return CLASS; }
}
class ConstantUTF8 extends Constant {
    String contents;
    ConstantUTF8(String contents) {
        this.contents = contents;
    }
    ConstantUTF8(DataInputStream di) 
        throws Exception {
        contents = di.readUTF();
    }

    private byte[] toJavaUTF8(String str) {
        byte[] utf8 = null;
        /* The stupid class file format encodes 0 different from UTF8,
         * so we must find zeroes and handles them ourselves. */
        int num_zeroes = 0;
        int index = 0;
        while ((index = contents.indexOf('\0', index)) > -1) {
            num_zeroes++;
            index++;
        }
        if (num_zeroes > 0) {
            int alloc_len = contents.length();
            utf8 = new byte[alloc_len];
            index = 0;
            int utf8index = 0; // Points at next spot to fill in utf8
            int oldindex = 0;
            while ((index = contents.indexOf('\0', index)) > -1) {
                byte[] utf8tmp;
                try {
                    utf8tmp = contents.substring(oldindex, index).
                        getBytes("UTF8");
                } catch (UnsupportedEncodingException e) {
                    throw new Error("UTF8 encoding not supported.  No, really.  Look: "+e);
                }
                int fulllength = utf8tmp.length + utf8index + 2;
                if (fulllength > alloc_len) {
                    alloc_len = Math.max(alloc_len*2, fulllength);
                    byte[] utf8new = new byte[alloc_len];
                    System.arraycopy(utf8, 0, utf8new, 0, utf8index);
                    utf8 = utf8new;
                }
                System.arraycopy(utf8tmp, 0, utf8, utf8index, utf8tmp.length);
                utf8index += utf8tmp.length;
                utf8[utf8index++] = (byte)0xc0;
                utf8[utf8index++] = (byte)0x80;
                index++;
                oldindex = index;
            }
            if (oldindex < contents.length()) {
                byte[] utf8tmp;
                try {
                    utf8tmp = contents.substring(oldindex).getBytes("UTF8");
                } catch (UnsupportedEncodingException e) {
                    throw new Error("UTF8 encoding not supported.  No, really.  Look: "+e);
                }
                int fulllength = utf8tmp.length + utf8index;
                if (fulllength > alloc_len) {
                    alloc_len = Math.max(alloc_len*2, fulllength);
                    byte[] utf8new = new byte[alloc_len];
                    System.arraycopy(utf8, 0, utf8new, 0, utf8index);
                    utf8 = utf8new;
                }
                System.arraycopy(utf8tmp, 0, utf8, utf8index, utf8tmp.length);
                utf8index += utf8tmp.length;
            }
            if (utf8index < alloc_len) {
                byte[] utf8new = new byte[utf8index];
                System.arraycopy(utf8, 0, utf8new, 0, utf8index);
                utf8 = utf8new;
            }
        } else {
            try {
                utf8 = contents.getBytes("UTF8");
            } catch (UnsupportedEncodingException e) {
                throw new Error("UTF8 encoding not supported.  No, really.  Look: "+e);
            }
        }
        return utf8;
    }

    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        byte[] utf8 = toJavaUTF8(contents);
        o.writeShort(utf8.length);
        o.write(utf8);
    }
    public int hashCode() {
        return contents.hashCode()^0x2a071e4b; // Random integer
    }
    public String toString() {
        return contents;
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantUTF8)) return false;
        ConstantUTF8 cc = (ConstantUTF8)other;
        return contents.equals(cc.contents);
    }
    public int constantNumber() { return UTF8; }

}
class ConstantString extends Constant {
    String contents;
    Constant utf8_constant;
    int indexbyte;
    ConstantString(String contents) {
        this.contents = contents;
    }
    ConstantString(DataInputStream di) 
        throws Exception {
        indexbyte = di.readUnsignedShort();
    }
    void resolve(Constant[] constants) {
        utf8_constant = constants[indexbyte];
        utf8_constant.resolve(constants);
        contents = utf8_constant.toString();
    }
    public void onInsertion(ConstantPool c) {
        utf8_constant = c.addUTF8(contents);
    }
    public boolean isLdcConstant() { return true; }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeShort(utf8_constant.index);
    }
    public int hashCode() {
        return contents.hashCode()^0x0ae21b47; // Random integer
    }
    public String toString() {
        return "\""+contents+"\"";
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantString)) return false;
        ConstantString cc = (ConstantString)other;
        return contents.equals(cc.contents);
    }
    public int constantNumber() { return STRING; }
}
abstract class ConstantRef extends Constant {
    String class_name;
    String name;
    String type;
    private ConstantClass classOf;
    ConstantNameAndType nameAndType;
    int indexbyte1, indexbyte2;
    ConstantRef(String classOf, String name, String type) {
        this.class_name = classOf;
        this.name = name;
        this.type = type;
    }
    public ConstantClass getClassOf() {
        return classOf;
    }
    ConstantRef(DataInputStream di) 
        throws Exception {
        indexbyte1 = di.readUnsignedShort();
        indexbyte2 = di.readUnsignedShort();
    }
    void resolve(Constant[] constants) {
        classOf = (ConstantClass)constants[indexbyte1];
        nameAndType = (ConstantNameAndType)constants[indexbyte2];
        classOf.resolve(constants);
        nameAndType.resolve(constants);
        class_name = classOf.getName();
        name = nameAndType.name;
        type = nameAndType.type;
    }
    public void setClassOf(ConstantClass classOf) {
        this.classOf = classOf;
    }
    public boolean isLdcConstant() { return true; }
    public void onInsertion(ConstantPool c) {
        setClassOf((ConstantClass)c.addClass(class_name));
        nameAndType = (ConstantNameAndType)c.addNameAndType(name, type);
    }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeShort(getClassOf().index);
        o.writeShort(nameAndType.index);
    }
    public int hashCode() {
        return class_name.hashCode()^name.hashCode()^type.hashCode()^0x0ae21b47; // Random integer
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantRef)) return false;
        ConstantRef cc = (ConstantRef)other;
        return class_name.equals(cc.class_name)
            && name.equals(cc.name)
            && type.equals(cc.type);
    }
}
class ConstantFieldRef extends ConstantRef {
    ConstantFieldRef(String classOf, String name, String type) {
        super(classOf, name, type);
        if (classOf == null) throw new Error("Null class for "+name);
    }
    ConstantFieldRef(FieldInfo f) {
        super(f.getClassInfo().getName(), f.getName(), 
              f.type.bytecodeString());
        if (f.getClassInfo() == null) throw new Error("Null class for "+name);
    }
    ConstantFieldRef(DataInputStream di) 
        throws Exception {
        super(di);
    }
    public String toString() {
        return "ConstantFieldRef "+getClassOf()+"."+name+"("+super.toString()+")";
    }
    public int constantNumber() { return FIELDREF; }
}
class ConstantMethodRef extends ConstantRef {
    MethodType type;
    ConstantMethodRef(String classOf, String name, MethodType type) {
        super(classOf, name, type.bytecodeString());
        this.type = type;
    }
    ConstantMethodRef(MethodInfo m) {
        this(m.getClassInfo().getName(), m.getName(),
             m.type);
    }
    ConstantMethodRef(DataInputStream di) 
        throws Exception {
        super(di);
    }
    public String toString() {
        return getClassOf()+"."+name+":"+type;
    }
    public int constantNumber() { return METHODREF; }
}
class ConstantInterfaceMethodRef extends ConstantMethodRef {
    ConstantInterfaceMethodRef(String classOf, String name, MethodType type) {
        super(classOf, name, type);
    }
    ConstantInterfaceMethodRef(MethodInfo m) {
        super(m);
    }
    ConstantInterfaceMethodRef(DataInputStream di) 
        throws Exception {
        super(di);
    }
    public int numArgs() {
        int numargs = 1;
        for (int i = 0; i < type.arguments.length; i++) 
            numargs+=type.arguments[i].wordSize();
        return numargs;
    }
    public int constantNumber() { return INTERFACE; }
}
class ConstantNameAndType extends Constant {
    String name;
    String type;
    ConstantUTF8 name_constant, type_constant;
    int indexbyte1, indexbyte2;
    ConstantNameAndType(String name, String type) {
        this.name = name;
        this.type = type;
    }
    ConstantNameAndType(DataInputStream di) 
        throws Exception {
        indexbyte1 = di.readUnsignedShort();
        indexbyte2 = di.readUnsignedShort();
    }
    void resolve(Constant[] constants) {
        name_constant = (ConstantUTF8)constants[indexbyte1];
        name_constant.resolve(constants);
        name = name_constant.toString();
        type_constant = (ConstantUTF8)constants[indexbyte2];
        name_constant.resolve(constants);
        type = type_constant.toString();
    }
    public void onInsertion(ConstantPool c) {
        name_constant = (ConstantUTF8)c.addUTF8(name);
        type_constant = (ConstantUTF8)c.addUTF8(type);
    }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeShort(name_constant.index);
        o.writeShort(type_constant.index);
    }
    public int hashCode() {
        return name.hashCode()^type.hashCode()^0x0ae21b47; // Random integer
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantNameAndType)) return false;
        ConstantNameAndType cc = (ConstantNameAndType)other;
        return name.equals(cc.name) && type.equals(cc.type);
    }
    public String toString() {
        return name+"&"+type+"("+name_constant.index+","+type_constant.index+")";
    }
    public int constantNumber() { return NAMEANDTYPE; }
}
class ConstantInteger extends Constant {
    int contents;
    ConstantInteger(int contents) {
        this.contents = contents;
    }
    ConstantInteger(DataInputStream di) throws Exception {
        contents = di.readInt();
    }
    public boolean isLdcConstant() { return true; }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeInt(contents);
    }
    public int hashCode() {
        return contents^0x0ae21b47; // Random integer
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantInteger)) return false;
        ConstantInteger cc = (ConstantInteger)other;
        return contents == cc.contents;
    }
    public int constantNumber() { return INTEGER; }

}
class ConstantFloat extends Constant {
    float contents;
    ConstantFloat(float contents) {
        this.contents = contents;
    }
    ConstantFloat(DataInputStream di) throws Exception {
        contents = di.readFloat();
    }
    public boolean isLdcConstant() { return true; }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeFloat(contents);
    }
    public int hashCode() {
        return Float.floatToIntBits(contents)^0x52e14ba7; // Random integer
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantFloat)) return false;
        ConstantFloat cc = (ConstantFloat)other;
        return contents == cc.contents;
    }
    public int constantNumber() { return FLOAT; }

}
class ConstantLong extends Constant {
    long contents;
    ConstantLong(long contents) {
        this.contents = contents;
    }
    ConstantLong(DataInputStream di) throws Exception {
        contents = di.readLong();
    }
    public boolean isLdcConstant() { return true; }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeLong(contents);
    }
    public boolean isDoubleConstant() { return true; }
    public int hashCode() {
        return ((int)contents)^0x3ea124b7; // Random integer
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantLong)) return false;
        ConstantLong cc = (ConstantLong)other;
        return contents == cc.contents;
    }
    public int constantNumber() { return LONG; }

}
class ConstantDouble extends Constant {
    double contents;
    ConstantDouble(double contents) {
        this.contents = contents;
    }
    ConstantDouble(DataInputStream di) throws Exception {
        contents = di.readDouble();
    }
    public boolean isLdcConstant() { return true; }
    void emitConstant(DataOutputStream o) throws IOException {
        super.emitConstant(o);
        o.writeDouble(contents);
    }
    public boolean isDoubleConstant() { return true; }
    public int hashCode() {
        return ((int)Double.doubleToLongBits(contents))^0x2c41aeb7;
    }
    public boolean equals(Object other) {
        if (!(other instanceof ConstantDouble)) return false;
        ConstantDouble cc = (ConstantDouble)other;
        return contents == cc.contents;
    }
    public int constantNumber() { return DOUBLE; }
}

/* End of class Constant*/
