/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * MethodInfo.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import uiuc.Jumbo.Util.*;

/**
 * A class describing a field of a java class
 * @version      1.0,  2 Oct 2001
 * @author       Lars Clausen
 *
*/

public class MethodInfo implements ClassMember {
    int flags;
    private String name;
    MethodType type;
    MonoList attributes = new ArrayMonoList();
    ConstantMethodRef methodconstant;
    ClassInfo inClass;

    public MethodInfo(ClassInfo inClass, int flags, String name, MethodType type) {
        this.flags = flags;
        if (name == null || name.equals("")) name = "<init>";
        this.name = name;
        this.type = type;
        this.inClass = inClass;
    }

    public MethodInfo(ClassInfo inClass, Method m) {
        flags = m.getModifiers();
        if (name == null || name.equals("")) name = "<init>";
        name = m.getName();
        type = new MethodType(m.getReturnType(), m.getParameterTypes());
        this.inClass = inClass;
    }
   
    public MethodInfo(ClassInfo inClass, Constructor m) {
        flags = m.getModifiers();
        name = "<init>";
        type = new MethodType(Void.TYPE, m.getParameterTypes());
        this.inClass = inClass;
    }
   
    public MethodInfo(DataInputStream di, ConstantPool pool) 
        throws Exception {
        flags = di.readUnsignedShort();
        int cindex = di.readUnsignedShort();
        if (cindex < 1 || cindex >= pool.size()) {
            throw new Error("Illegal method name constant "+cindex);
        } 
        Constant n = pool.get(cindex);
        if (!(n instanceof ConstantUTF8)) {
            throw new Error("Non-UTF8 method name constant "+n);
        } 
        name = n.toString();
        if (name == null || name.equals("")) name = "<init>";

        cindex = di.readUnsignedShort();
        if (cindex < 1 || cindex >= pool.size()) {
            throw new Error("Illegal method signature constant "+cindex);
        } 
        Constant s = pool.get(cindex);
        if (!(s instanceof ConstantUTF8)) {
            throw new Error("Non-UTF8 method signature constant "+s);
        }
        type = MethodType.parseType(s.toString());
        if (type == null) {
            throw new Error("Illegal method signature "+s.toString());
        }

        int count = di.readUnsignedShort();
        for (int i = 0; i < count; i++) {
            int nameindex = di.readUnsignedShort();
            int length = di.readInt();
            String name = pool.get(nameindex).toString();
            di.skipBytes(length);
        }
    }
   
    public Constant getConstant() { return methodconstant; }
    public void setConstant(Constant c) { methodconstant = (ConstantMethodRef)c; }

    public ClassInfo getClassInfo() { return inClass; }
    public String getName() { 
        return name; 
    }
    public MethodType getType() { return type; }
    public Type getReturnType() { return type.returntype; }

    public boolean flagsSet(int flags) {
        return (this.flags & flags) == flags;
    }

    public void addCode(Instr code) {
        CodeAttribute ca = new CodeAttribute(this);
        ca.code = code;
        if (code.handlers != null)
            ca.exceptions = code.handlers;
        attributes.add(ca);
    }

    public boolean isStatic() {
        return (flags & Flags.STATIC) != 0;
    }

    /** Returns the method with the more specific type.
     * Note: Doesn't handle ambiguous methods (e.g. foo(int, long) vs
     * foo(long, int)).
     * @param m A MethodInfo instance, or null.
     */
    public MethodInfo findBestMethod(MethodInfo m) {
        if (m == null) return this;
        Type[] t1 = type.arguments;
        Type[] t2 = m.type.arguments;
        MethodInfo better = null;
        
        for (int i = t1.length-1; i >= 0; i--) {
            if (!t1[i].conformsMethod(t2[i])) {
                return m;
            }
            if (!t2[i].conformsMethod(t1[i])) {
                return this;
            }
        }
        return this;
    }

    public String toString() {
        return type.toString(inClass.getName()+"."+name);
    }

    public void prepareForEmit(ConstantPool c) {
        if (inClass.isInterface())
            methodconstant = (ConstantInterfaceMethodRef)c.addInterface(this);
        else
            methodconstant = (ConstantMethodRef)c.addMethod(this);
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.prepareForEmit(c);
        }
    }

    public void emitTo(DataOutputStream o) throws IOException {
        Constant nameconstant = methodconstant.nameAndType.name_constant;
        Constant typeconstant = methodconstant.nameAndType.type_constant;
        o.writeShort(flags);
        nameconstant.emitIndex(o);
        typeconstant.emitIndex(o);

        o.writeShort(attributes.size());
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.emitTo(o);
        }
    }

    public void getAdded(ClassInfo c) {
        c.addMember(this);
    }

    public boolean equals(Object o) {
        if (!(o instanceof MethodInfo)) return false;
        MethodInfo m = (MethodInfo)o;
        if (flags != m.flags) return false;
        if (!name.equals(m.name)) return false;
        if (!type.equals(m.type)) return false;
        if (!inClass.equals(m.inClass)) return false;
        if (attributes.size() != m.attributes.size()) return false;
        for (Iterator i1 = attributes.iterator(),
                 i2 = m.attributes.iterator(); i1.hasNext(); ) {
            if (!i1.next().equals(i2.next())) return false;
        }
        return true;
    }
}

/* End of class MethodInfo*/
