/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * MethodType.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;

/**
 * This class handles the type of a method
 * @version      1.0,  5 Oct 2001
 * @author       Lars Clausen
 *
*/

public class MethodType {
    public Type[] arguments;
    public Type returntype;
    public MethodType(Type returntype, Type[] args) {
        this.returntype = returntype;
        this.arguments = args;
    }
    /** Convenience constructor for zero-argument methods */
    public MethodType(Type returntype) {
        this(returntype, new Type[0]);
    }
    /** Convenience constructor for oneo-argument methods */
    public MethodType(Type returntype, Type arg1) {
        this(returntype,
             new Type[] { arg1 });
    }
    /** Convenience constructor for two-argument methods */
    public MethodType(Type returntype, Type arg1, Type arg2) {
        this(returntype,
             new Type[] { arg1, arg2 });
    }
    /** Convenience constructor for three-argument methods */
    public MethodType(Type returntype, Type arg1, Type arg2, Type arg3) {
        this(returntype,
             new Type[] { arg1, arg2, arg3 });
    }

    /** Constructor for methods found with java.reflect */
    public MethodType(Class returntype, Class[] formaltypes) {
        this.returntype = Type.convertType(returntype);
        arguments = new Type[formaltypes.length];
        for (int i = 0; i < formaltypes.length; i++) {
            arguments[i] = Type.convertType(formaltypes[i]);
        }
    }

    /** Parses a method signature to a type.
     * A method signature is of the form (X*)Y, where X is
     * a parameter signature and Y is the return signature.
     */
    private static final Type[] typer = new Type[0];
    public static MethodType parseType(String s) {
        try {
        if (s.charAt(0) != '(') return null;
        int i = 1;
        List args = new ArrayList();
        while (s.charAt(i) != ')') {
            args.add(Type.parseBytecodeType(s, i));
            i = Type.bytecodeTypeLength(s, i);
        }
        return new MethodType(Type.parseBytecodeType(s, i+1),
                              (Type[])args.toArray(typer));
        } catch (StringIndexOutOfBoundsException e) {
            e.printStackTrace();
            throw e;
        }
    }

    public boolean conforms(MethodType m) {
        if (m.arguments.length != arguments.length) return false;
        for (int i = 0; i < arguments.length; i++) {
            if (!(arguments[i].conformsMethod(m.arguments[i])))
                return false;
        }
        return true;
    }

    public String bytecodeString() {
        StringBuffer buf = new StringBuffer("(");
        for (int i = 0; i < arguments.length; i++) {
            buf.append(arguments[i].bytecodeString());
        }
        buf.append(")");
        buf.append(returntype.bytecodeString());
        return buf.toString();
    }

    public String toString() {
        return toString("");
    }

    public String toString(String methodname) {
        StringBuffer buf = new StringBuffer();
        buf.append(returntype.toString());
        buf.append(" ");
        buf.append(methodname);
        buf.append("(");
        for (int i = 0; i < arguments.length; i++) {
            if (i > 0) buf.append(", ");
            buf.append(arguments[i].toString());
        }
        buf.append(")");
        return buf.toString();
        
    }

    public boolean equals(Object o) {
        if (!(o instanceof MethodType)) return false;
        MethodType m = (MethodType)o;
        if (!returntype.equals(m.returntype)) return false;
        if (arguments.length != m.arguments.length) return false;
        for (int i = 0; i < arguments.length; i++) {
            if (!arguments[i].equals(m.arguments[i])) return false;
        }
        return true;
    }

    /** Unit tests */
    public static void unitTest() {
        Test.assertEquals(new MethodType(Type.int_type, Type.int_type).bytecodeString(),
                          "(I)I");
        Test.assertEquals(new MethodType(Type.void_type, Type.long_type, 
                                         new Type("java/lang/Object"),
                                         Type.boolean_type).bytecodeString(),
                          "(JLjava/lang/Object;Z)V");
    }
}

/* End of class MethodType*/
