/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import uiuc.Jumbo.Util.*;
import java.util.*;

/**
 * Represents an anonymous class instance creation.
 *
 * @author      Lars Clausen
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class AnonymousClassCreation extends ClassInstanceCreation 
    implements ClassDefiner {
   /**
    * @param c  name of the class being instanced
    * @param a  optional argument list
    */
   public AnonymousClassCreation( Identifier c, List a, ClassBody b ) { 
       super(c, a);
       setBody(b);
   }

   protected boolean replaceUC(Node old_child, Node new_child) {
       setBody((ClassBody)new_child);
       return true;
   }

   private void setBody(ClassBody b) {
       body = b;
       setParent(b);
   }

    public IR toIR() {
        return IR.make("anonymousClass", new ArrayMonoList()
                       .add(IR.make(class_name.toString()))
                       .add(toIRList(arguments))
                       .add(body.toIR()));
    }

    public void toSource() {
       print("new ");
       class_name.toSource();
       print("(");
       printSourceList(arguments, ", ");
       print(") ");
       ocurly();
       linebreak();
       body.toSource();
       ccurly();
   }

    /** Returns the variable defining the given field, if any.
     */
    public Variable findField(String field) {
        for (Iterator i = body.members.iterator();
             i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof FieldDeclaration) {
                FieldDeclaration fd = (FieldDeclaration)o;
                Variable v = fd.findField(field);
                if (v != null) return v;
            }
        }
        return null;
    }

    /** Returns the type of the given field, if any.
     */
    public Type findFieldType(String field) {
        for (Iterator i = body.members.iterator();
             i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof FieldDeclaration) {
                FieldDeclaration fd = (FieldDeclaration)o;
                Variable v = fd.findField(field);
                if (v != null) return fd.type;
            }
        }
        return null;
    }

    /** Find a method based on name and arguments.
     * @param name The name of the method.
     * @param args A List of Types indicating the types of the actual 
     * arguments.
     * @param superfinder An instance of Type.SuperFinder allowing lookup
     * of superclasses.
     * @return A matching method, or null if none was found.
     */
    public MethodDeclaration findMethod(String name, List args,
                                        TypeEnv typeenv) {
        METHODS:
        for (Iterator i = body.members.iterator();
             i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof MethodDeclaration) {
                MethodDeclaration md = (MethodDeclaration)o;
                if (md.conformsTo(name, args, typeenv))
                    return md;
            }
        }
        if (!name.equals("<init>")) {
            System.err.println("Can't look up methods in superclasses yet");
        }
        return null;
    }

    /** Find the return type of a method based on name and arguments.
     * @param name The name of the method.
     * @param args A List of Types indicating the types of the actual 
     * arguments.
     * @param superfinder An instance of Type.SuperFinder allowing lookup
     * of superclasses.
     * @return A matching method, or null if none was found.
     */
    public Type findMethodType(String name, List args, TypeEnv typeenv) {
        METHODS:
        for (Iterator i = body.members.iterator();
             i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof MethodDeclaration) {
                MethodDeclaration md = (MethodDeclaration)o;
                if (md.conformsTo(name, args, typeenv))
                    return md.type;
            }
        }
        if (!name.equals("<init>")) {
            System.err.println("Can't look up methods in superclasses yet");
        }
        return null;
    }

    /** Returns an identifier for the name of the super class */
    public String getSuperName() {
        if (class_name == null) return "java.lang.Object";
        return class_name.toString();
    }

    /** Returns the name of this class. */
    public String getName() {
        throw new Error("Anon classes have no name");
    }

    /** Returns the fully qualified class name of this class. */
    public String getFQCN() {
        throw new Error("Anon classes have no name");
    }

    /** Returns a list of names of interfaces */
    public List getInterfaces() {
        return new ArrayList(0);
    }

    /** Returns true if this class is final.  It always is. */
    public boolean isFinal() {
        return true;
    }

    public String toString() {
        return "new "+class_name.toString()+"() { <"+(body != null?body.members.size():0)+" members> }";
    }

    /**
     * Contents of the class body
     */
    protected ClassBody body;
}
