/* Copyright (c) 2002, University of Illinois.  All rights reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * ClassWrapper.java
 *
 */

package uiuc.Jumbo.Ast;

import java.lang.reflect.*;
import java.util.*;

/**
 * A ClassDefiner wrapper around the Class object (for sourceless classes)
 * @version      1.0, 22 Nov 2002
 * @author       Lars Clausen
 *
*/

public class ClassWrapper implements ClassDefiner {
    public ClassWrapper(Class c) {
        this.thisclass = c;
    }

    /** Find an AST node representing this field, if possible.
     */
    public Variable findField(String name) {
        System.err.println("Can't find AST node for field "+name+" on compiled class "+this);
        return null;
    }
    /** Find an AST node representing this method, if possible.
     */
    public MethodDeclaration findMethod(String name, List args,
                                 TypeEnv typeenv) {
        throw new Error("Can't find AST node for method "+name+" on compiled class "+this);
    }

    /** Return the type of the named field in this class.
     */
    public Type findFieldType(String name) {
        try {
            Field f = thisclass.getDeclaredField(name);
            return new Type(f.getType());
        } catch (NoSuchFieldException e) {
            Class sc = thisclass.getSuperclass();
            if (sc != null)
                return new ClassWrapper(sc).findFieldType(name);
            return null;
        }
    }

    private Type findMethodTypeInternal(String name, List args, TypeEnv typeenv) {
        try {
            Method m = thisclass.getDeclaredMethod(name, 
                                                   (Class[])args.toArray(new Class[0]));
            return new Type(m.getReturnType());
        } catch (NoSuchMethodException e) {
            Class sc = thisclass.getSuperclass();
            if (sc != null) 
                return new ClassWrapper(sc).
                    findMethodTypeInternal(name, args, typeenv);
            throw new Error("Can't find type for "+this+"."+name+"("+args+")");
        }
    }

    /** Return the return type of the indicated method in this class.
     */
    public Type findMethodType(String name, List args, TypeEnv typeenv) {
        List l = new ArrayList(args.size());
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            Type t = (Type)i.next();
            l.add(t.toClass());
        }
        return findMethodTypeInternal(name, l, typeenv);
    }
    /** Return the FQCN of the super class.
     */
    public String getSuperName() {
        Class sc = thisclass.getSuperclass();
        if (sc != null) return sc.getName();
        return null;

    }
    /** Return the name of this class.
     */
    public String getName() {
        return thisclass.getName();
    }
    /** Returns the fully qualified class name of this class.
     * For ClassWrapper, the name is always fully qualified.
     */
    public String getFQCN() {
        return thisclass.getName();
    }

    /** Returns a list of names of interfaces */
    public List getInterfaces() {
        Class[] interfaces = thisclass.getInterfaces();
        List l = new ArrayList(interfaces.length);
        for (int i = 0; i < interfaces.length; i++) {
            l.add(interfaces[i].getName());
        }
        return l;
    }

    /** Return true if this is a final class.
     */
    public boolean isFinal() {
        int mods = thisclass.getModifiers();
        return Modifier.isFinal(mods);
    }

    Class thisclass;
}

/* End of class ClassWrapper*/
