/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * IR.java
 *
 *
 */

package uiuc.Jumbo.Ast;

import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;
import java.lang.reflect.*;
import java.util.*;
import uiuc.Jumbo.Jaemus.*;
import java.io.StringWriter;

/**
 * A simple intermediate form that is easy to output and lift.
 * @version      1.0, 17 Oct 2002
 * @author       Lars Clausen
 *
*/

public abstract class IR {
    public static IRInt make(int i) {
        return new IRInt(i);
    }
    public static IRBoolean make(boolean i) {
        return new IRBoolean(i);
    }
    public static IRLong make(long i) {
        return new IRLong(i);
    }
    public static IRFloat make(float i) {
        return new IRFloat(i);
    }
    public static IRDouble make(double i) {
        return new IRDouble(i);
    }
    public static IRCharacter make(char i) {
        return new IRCharacter(i);
    }
    public static IRString make(String i) {
        return new IRString(i);
    }
    /** Make a Code block */
    public static IRCode make(String constructor, MonoList args) {
        return new IRCode(constructor, args);
    }
    /** Make a MonoList of something */
    public static IRList make(MonoList list) {
        return new IRList(list);
    }
    /** Make a MonoList of something */
    public static IRCaseList makeCaseList(MonoList list) {
        return new IRCaseList(list);
    }
    /** Make a case block for a switch statement
     * @param list A list of IRCodes: The switch labels
     * @param block An IRCode: The block executed for these labels
     */
    public static IRCase make(MonoList list, IRCode block) {
        return new IRCase(list, block);
    }
    /** Make an exception handling block.
     * @param flags May be 0 or Flags.FINAL
     * @param t The type of the exception handled
     * @param id The name of the exception variable
     * @param block The IRCode block for the handler
     */
    public static IRCatch make(int flags, uiuc.Jumbo.Jaemus.Type t, IR id,
                               IRCode block) {
        return new IRCatch(flags, t, id, block);
    }
    /** Make a type */
    public static IRType make(uiuc.Jumbo.Jaemus.Type t) {
        return new IRType(t);
    }
    /** Make an IR representing an antiquoted area */
    public static IRAST make(Node n) {
        return new IRAST(n);
    }
    public static IRArg make(int flags, uiuc.Jumbo.Jaemus.Type t, IR id) {
        return new IRArg(flags, t, id);
    }
    public static IRNull makeNull(String classname) {
        return new IRNull(classname);
    }
    public static IRCode makeNoop() {
        return make("noop", new ArrayMonoList());
    }

    /** Lifts this IR into a higher level of quoting (later evaluation).
     */
    public abstract IRCode lift();
    /** Returns a Class object representing the type of this IR 
     */
    public abstract Class toType();
    /** Returns an Object representing this IR passable to Method.invoke.
     */
    public abstract Object toObject();
    /** Returns a string of this code as Java source (lifted once).
     */
    public abstract String toSource();
    public static void main(String[] argv) {
        IRCode foo = (IRCode)make("binOp", new ArrayMonoList()
                                  .add(make(Instr.ADD))
                                  .add(make("integerConstant", new ArrayMonoList(make(2))))
                                  .add(make("integerConstant", new ArrayMonoList(make(3)))));
        System.out.println(foo);
        System.out.println(foo.toSource());
        System.out.println(foo.toObject());

        IRCode bar = (IRCode)make("defineVariable", new ArrayMonoList()
                                  .add(make(0))
                                  .add(make(uiuc.Jumbo.Jaemus.Type.int_type))
                                  .add(make("x"))
                                  .add(make("integerConstant", new ArrayMonoList(make(4)))));
        System.out.println(bar);
        System.out.println(bar.toSource());
        System.out.println(bar.toObject());

        IRCode baz = (IRCode)make("switchLabel", new ArrayMonoList()
                                  .add(make(""))
                                  .add(make("integerConstant", new ArrayMonoList(make(1))))
                                  .add(make(new ArrayMonoList()
                                            .add(make(new ArrayMonoList()
                                                      .add(make("integerConstant", new ArrayMonoList(make(2))))
                                                      .add(make("integerConstant", new ArrayMonoList(make(3)))),
                                                      make("breakStatement", new ArrayMonoList())))))
                                  .add(make("integerConstant", new ArrayMonoList(make(4)))));
        System.out.println(baz);
        System.out.println(baz.toSource());
        System.out.println(baz.toObject());

        IRCode qux = (IRCode)make("handlersWithFinally", new ArrayMonoList()
                                  .add(make("integerConstant", new ArrayMonoList(make(1))))
                                  .add(make(new ArrayMonoList()
                                            .add(make(0, new uiuc.Jumbo.Jaemus.Type("Exception"),
                                                      make("e"),
                                                      make("integerConstant",
                                                           new ArrayMonoList(make(3)))))))
                                  .add(make("integerConstant", new ArrayMonoList(make(5)))));
        System.out.println(qux);
        System.out.println(qux.toSource());
        System.out.println(qux.toObject());
    }
}

class IRInt extends IR {
    int value;
    public IRInt(int i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("integerConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Integer.TYPE;
    }
    public Object toObject() {
        return new Integer(value);
    }
    public String toString() {
        return Integer.toString(value);
    }
    public String toSource() {
        return toString();
    }
}

class IRBoolean extends IR {
    boolean value;
    public IRBoolean(boolean i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("booleanConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Boolean.TYPE;
    }
    public Object toObject() {
        return new Boolean(value);
    }
    public String toSource() {
        return toString();
    }
    public String toString() {
        return (value?"true":"false");
    }
}


class IRLong extends IR {
    long value;
    public IRLong(long i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("longConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Long.TYPE;
    }
    public Object toObject() {
        return new Long(value);
    }
    public String toSource() {
        return toString()+"l";
    }
    public String toString() {
        return Long.toString(value);
    }
}

class IRFloat extends IR {
    float value;
    public IRFloat(float i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("floatConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Float.TYPE;
    }
    public Object toObject() {
        return new Float(value);
    }
    public String toSource() {
        return toString()+"f";
    }
    public String toString() {
        return Float.toString(value);
    }
}

class IRDouble extends IR {
    double value;
    public IRDouble(double i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("doubleConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Double.TYPE;
    }
    public Object toObject() {
        return new Double(value);
    }
    public String toSource() {
        return toString()+"d";
    }
    public String toString() {
        return Double.toString(value);
    }
}

class IRCharacter extends IR {
    char value;
    public IRCharacter(char i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("characterConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return Character.TYPE;
    }
    public Object toObject() {
        return new Character(value);
    }
    public String toSource() {
        //XXX Needs quoting
        return "'"+value+"'";
    }
    public String toString() {
        return "'"+value+"'";
    }
}

class IRString extends IR {
    String value;
    public IRString(String i) {
        value = i;
    }
    public IRCode lift() {
        return new IRCode("stringConstant", new ArrayMonoList(this));
    }
    public Class toType() {
        return value.getClass();
    }
    public Object toObject() {
        return value;
    }
    public String toSource() {
        //XXX Needs quoting?
        return "\""+value+"\"";
    }
    public String toString() {
        return "\""+value+"\"";
    }
}

class IRCode extends IR {
    String constructor;
    MonoList args;
    public IRCode(String constructor, MonoList args) {
        this.constructor = constructor;
        this.args = args;
    }
    public IRCode addArg(IR arg) {
        args.add(arg);
        return this;
    }
    static Class constr;
    static {
        try {
            constr = Class.forName("uiuc.Jumbo.Compiler.Constructor");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find constructor class!");
        }
    }
    private Method getMethod(boolean try_singles) {
        Class[] types = new Class[args.size()];
        int index = 0;
        boolean has_singles = false;
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            // You cannot toType an IRAst - so if we find an Ast
            // while testing for singles, we have to fall back to lists.
            if (try_singles && (ir instanceof IRAST)) return null;
            if (try_singles && (ir instanceof IRList)) {
                IRList irlist = (IRList)ir;
                IR content = irlist.getSingle();
                if (content != null) {
                    types[index] = content.toType();
                    has_singles = true;
                } else
                    types[index] = ir.toType();
            } else {
                types[index] = ir.toType();
            }
            index++;
        }
        if (try_singles && !has_singles) return null;
        try {
            return constr.getMethod(constructor, types);    
        } catch (Exception e) {
            return null;
        }
    }
    public IRCode lift() {
        boolean use_singles = (getMethod(true) != null);
        IRList liftedargs = new IRList();
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            if (use_singles && ir instanceof IRList) {
                IRList irlist = (IRList)ir;
                IR content = irlist.getSingle();
                if (content != null)
                    liftedargs.add(content.lift());
                else
                    liftedargs.add(ir.lift());
            } else
                liftedargs.add(ir.lift());
        }
        IRCode invoke = new IRCode("invoke",
                                   new ArrayMonoList()
                                   .add(make("getQualifiedNameOrClass",
                                             new ArrayMonoList(make("Constructor"))))
                                   .add(make(constructor))
                                   .add(liftedargs));
        return invoke;
    }
    public Object toObject() {
        boolean use_singles = true;
        Method m = getMethod(use_singles);
        if (m == null) {
            use_singles = false;
            m = getMethod(use_singles);
            if (m == null) {
                throw new Error("Can't turn "+this+" into code");
            }
        }
        Object[] objects = new Object[args.size()];
        int index = 0;
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            if (use_singles && ir instanceof IRList) {
                IRList irlist = (IRList)ir;
                IR content = irlist.getSingle();
                if (content != null)
                    objects[index] = content.toObject();
                else
                    objects[index] = ir.toObject();
            } else
                objects[index] = ir.toObject();
            index++;
        }
        try {
            return (Code)m.invoke(null, objects);
        } catch (Exception e) {
            System.err.println("Can't turn "+this+" into code");
            e.printStackTrace();
            throw new Error("Code translation error: "+e);
        }
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Compiler.Code");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find Code class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public String toString() {
        String s = "("+constructor+", [";
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            s += i.next();
            if (i.hasNext()) s+= ", ";
        }
        s += "])";
        return s;
    }
    public String toSource() {
        boolean use_singles = (getMethod(true) != null);
        String s = "Constructor."+constructor+"(";
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            if (use_singles && ir instanceof IRList) {
                IRList irlist = (IRList)ir;
                IR content = irlist.getSingle();
                if (content != null)
                    s += content.toSource();
                else
                    s += ir.toSource();
            } else
                s += ir.toSource();
            if (i.hasNext()) s+= ", ";
        }
        s += ")";
        return s;
    }
}

class IRList extends IR {
    MonoList elements;
    public IRList() {
        elements = new ArrayMonoList();
    }
    public IRList(MonoList items) {
        elements = items;
    }
    public void add(IR el) {
        elements.add(el);
    }
    /** Returns the only element of a one-element list, or null if there are
     * more or less than one elements */
    public IR getSingle() {
        if (elements.size() == 1) return (IR)elements.get(0);
        return null;
    }
    public IRCode lift() {
        IRCode array = new IRCode("newObject",
                                  new ArrayMonoList()
                                  .add(make("ArrayMonoList"))
                                  .add(make(new ArrayMonoList())));
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            array = new IRCode("invoke", new ArrayMonoList()
                               .add(array)
                               .add(make("add"))
                               .add(make(new ArrayMonoList()
                                         .add(((IR)i.next()).lift()))));
        }
        return array;
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Util.MonoList");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find MonoList class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public Object toObject() {
        MonoList list = new ArrayMonoList();
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            list.add(((IR)i.next()).toObject());
        }
        return list;
    }
    public String toString() {
        String s = "[";
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            s += i.next();
            if (i.hasNext()) s+= ", ";
        }
        s += "]";
        return s;
    }
    public String toSource() {
        String s = "new ArrayMonoList()";
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            s += ".add("+((IR)i.next()).toSource()+")";
        }
        return s;
    }
    public IRList flatten() {
        MonoList l = new ArrayMonoList();
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof IRList) {
                l.addAll(((IRList)o).flatten().elements);
            } else {
                l.add(o);
            }
        }
        elements = l;
        return this;
    }
}

class IRCaseList extends IRList {
    public IRCaseList() {
        super();
    }
    public IRCaseList(MonoList items) {
        super(items);
    }
    public IRCode lift() {
        IRCode array = new IRCode("newObject",
                                  new ArrayMonoList()
                                  .add(make("ArrayMonoList"))
                                  .add(make(new ArrayMonoList())));
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            String add_method = "add";
            if (ir instanceof IRAST) add_method = "addAll";
            array = new IRCode("invoke", new ArrayMonoList()
                               .add(array)
                               .add(make(add_method))
                               .add(make(new ArrayMonoList()
                                         .add(ir.lift()))));
        }
        return array;
    }
    public String toSource() {
        String s = "new ArrayMonoList()";
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            IR ir = (IR)i.next();
            String add_method = ".add";
            if (ir instanceof IRAST) add_method = ".addAll";
            s += add_method+"("+ir.toSource()+")";
        }
        return s;
    }
}

class IRCase extends IR {
    IRList caselist;
    IRCode block;
    public IRCase(MonoList list, IRCode block) {
        this.caselist = make(list);
        this.block = block;
    }
    public IRCode lift() {
        return new IRCode("newObject", new ArrayMonoList()
                          .add(make("SwitchCase"))
                          .add(make(new ArrayMonoList()
                                    .add(caselist.lift())
                                    .add(block.lift()))));
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Compiler.SwitchCase");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find SwitchCase class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public Object toObject() {
        return new SwitchCase((MonoList)caselist.toObject(), (Code)block.toObject());
    }
    public String toString() {
        String s = "(";
        for (Iterator i = caselist.elements.iterator(); i.hasNext(); ) {
            s += "case "+((IR)i.next())+": ";
        }
        s += block;
        return s;
    }
    public String toSource() {
        String s = "new SwitchCase(";
        s += caselist.toSource();
        return s+", "+block.toSource()+")";
    }   
}
class IRCatch extends IR {
    int flags;
    uiuc.Jumbo.Jaemus.Type t;
    IR id;
    IRCode block;
    public IRCatch(int flags, uiuc.Jumbo.Jaemus.Type t, IR id, IRCode block) {
        this.flags = flags;
        this.t = t;
        this.id = id;
        this.block = block;
    }
    public IRCode lift() {
        return new IRCode("newObject", new ArrayMonoList()
                          .add(make("Handler"))
                          .add(make(new ArrayMonoList()
                                    .add(make(flags).lift())
                                    .add(make(t).lift())
                                    .add(id.lift())
                                    .add(block.lift()))));
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Compiler.Handler");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find Handler class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public Object toObject() {
        return new Handler(flags, t, (String)id.toObject(), (Code)block.toObject());
    }
    public String toString() {
        return "(catch ("+flags+" "+t+" "+id+") "+block+")";
    }
    public String toSource() {
        return "new Handler("+flags+", "+new IRType(t).toSource()+
            ", \""+id.toSource()+"\", "+block.toSource()+")";
    }
}
class IRType extends IR {
    uiuc.Jumbo.Jaemus.Type t;
    public IRType(uiuc.Jumbo.Jaemus.Type t) {
        this.t = t;
    }
    private IRCode liftBaseType(uiuc.Jumbo.Jaemus.Type t) {
        if (t.isReference()) {
            return new IRCode("newObject", new ArrayMonoList()
                              .add(make("Type"))
                              .add(make(new ArrayMonoList()
                                        .add(make(t.getClassName()).lift()))));
        } else {
            return new IRCode("getField", new ArrayMonoList()
                              .add(make("getQualifiedNameOrClass", 
                                        new ArrayMonoList(make("Type"))))
                              .add(make(t+"_type")));
        }
    }
    private IRCode liftType(uiuc.Jumbo.Jaemus.Type t) {
        if (t.isArray()) {
            IRCode base = liftType(t.array_contents);
            return new IRCode("newObject", new ArrayMonoList()
                              .add(make("Type"))
                              .add(make(new ArrayMonoList()
                                        .add(base))));
        } else {
            return liftBaseType(t);
        }
    }
    public IRCode lift() {
        return liftType(t);
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Jaemus.Type");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find Type class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public Object toObject() {
        return t;
    }
    public String toString() {
        return t.toString();
    }
    private String toSourceBaseType(uiuc.Jumbo.Jaemus.Type t) {
        if (t.isReference()) {
            return "new Type(\""+t.getClassName()+"\")";
        } else {
            return "Type."+t+"_type";
        }
    }
    private String toSourceType(uiuc.Jumbo.Jaemus.Type t) {
        if (t.isArray()) {
            return "new Type("+toSourceType(t.array_contents)+")";
        } else {
            return toSourceBaseType(t);
        }
    }
    public String toSource() {
        return toSourceType(t);
    }
}

/** This class encodes the level below the lowest level of IR */
class IRAST extends IR {
    Node n;
    public IRAST(Node n) {
        this.n = n;
    }

    public IRCode lift() {
        return (IRCode)n.toIR();
    }

    public Object toObject() {
        throw new Error("Can't toObject IRAST");
    }

    public String toString() {
        StringWriter s = new StringWriter();
        Node.setOutputToString(s);
        n.toSource();
        Node.endOutput();
        return s.toString();
    }

    public String toSource() {
        return toString();
    }

    public Class toType() {
        throw new Error("Can't toType IRAST?");
    }
}

class IRArg extends IR {
    int flags;
    uiuc.Jumbo.Jaemus.Type t;
    IR id;
    public IRArg(int flags, uiuc.Jumbo.Jaemus.Type t, IR id) {
        this.flags = flags;
        this.t = t;
        this.id = id;
    }
    public IRCode lift() {
        return new IRCode("newObject", new ArrayMonoList()
                          .add(make("ArgDef"))
                          .add(make(new ArrayMonoList()
                                    .add(make(flags).lift())
                                    .add(make(t).lift())
                                    .add(id.lift()))));
    }
    private static Class type;
    { 
        try {
            type = Class.forName("uiuc.Jumbo.Compiler.ArgDef");
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find ArgDef class: "+e);
        }
    }
    public Class toType() {
        return type;
    }
    public Object toObject() {
        return new ArgDef(flags, t, (String)id.toObject());
    }
    public String toString() {
        return flags+" "+t+" "+id;
    }
    public String toSource() {
        return "new ArgDef("+flags+", "+new IRType(t).toSource()+
            ", "+id.toSource()+")";
    }
}

class IRNull extends IR {
    Class nulltype;
    public IRNull(String classname) {
        try {
            this.nulltype = Class.forName(classname);
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find class for null type: "+e);
        }
    }
    public IRCode lift() {
        return new IRCode("nullConstant", new ArrayMonoList());
    }
    public Class toType() {
        return nulltype;
    }
    public Object toObject() {
        return null;
    }
    public String toString() {
        return "null";
    }
    public String toSource() {
        return "null";
    }
}

/* End of class IR*/
