/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/**
 * Represents a normal method declaration.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;

import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;

public class MethodDeclaration extends Node implements ClassBodyStatement {
    /**
     * Creates a MethodDeclaration with a name and parameter list only.
     *
     * @param n name of the method.
     * @param p optional parameters of the method.
     */
    public MethodDeclaration( Identifier n, LinkedList p ) { 
        setName(n);
        parameters = p;
        setParents(parameters);
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
        if (old_child == name) {
            setName((Identifier)new_child);
            return true;
        }
        else if (old_child == type) {
            setType((Type)new_child);
            return true;
        }
        else if (old_child == body) {
            setBody((StatementList)new_child);
            return true;
        }
        else if (replaceInList(parameters, old_child, new_child)) {
            return true;
        }
        else if (replaceInList(thrown_classes, old_child, new_child)) {
            return true;
        }
        else {
            return false;
        }
    }

    private void setName(Identifier n) {
        name = n;
        setParent(name);
    }

    public IR toIR() {
        IRCode ir = IR.make("defineMethod", new ArrayMonoList());
        ir.addArg(modifiers != null?modifiers.toIR():IR.make(0));
        ir.addArg(type != null?type.toIR():IR.make(uiuc.Jumbo.Jaemus.Type.void_type));
        ir.addArg(name.toIR());
        ir.addArg(toIRList(parameters));
        ir.addArg(toIRList(thrown_classes));
        if (body != null)
            ir.addArg(body.toIR());
        else
            ir.addArg(IR.make("statements", new ArrayMonoList(IR.make(new ArrayMonoList()))));
        return ir;
    }

    /**
     * Output source for this method declaration
     */
    public void toSource() {
        if ( modifiers != null ) { 
            modifiers.toSource();
        }

        type.toSource();
        print(" ");

        name.toSource();

        oparen();
        printSourceList( parameters, ", " );
        cparen();

        if (thrown_classes != null && thrown_classes.size() > 0) {
            linebreak();
            print("throws ");
            printSourceList( thrown_classes, ", " );
        }
        if (parent instanceof InterfaceBody) {
            print(";");
            linebreak();
        }
        else {
            print(" ");
            if ( body != null && body.size() > 0) {
                body.toSource();
            } else {
                if (isAbstract())
                    print(";");
                else
                    print("{}");
            }
            linebreak();
        }
    }

    public String toString() {
        String result;
        if (name != null) 
            result = type.toString()+" "+name;
        else
            result = ((ClassDeclaration)parent.parent).name.toString();
        result += "(";
        if (parameters != null)
            for (Iterator i = parameters.iterator(); i.hasNext(); ) {
                result += i.next().toString();
                if (i.hasNext()) result += ", ";
            }
        result += ")";
        return result;
    }

    /**
     * @param m list of modifiers
     */
    public void setModifiers( ModifierList m ) {
        modifiers = m;
    }

    /**
     * @param t type of method
     */
    public void setType( Type t ) {
        type = t;
        setParent(t);
    }

    /**
     * @param thl throws list of the method
     */
    public void setThrowsList( LinkedList thl ) {
        thrown_classes = thl;
    }

    /**
     * @param b body of the method
     */
    public void setBody( StatementList s ) {
        body = s;
        setParent(s);
    }

    /** Returns true if this method conforms to the given name and args */
    public boolean conformsTo(String name, List args, TypeEnv typeenv) {
        if ((this.name == null && name.equals("<init>")) ||
                (this.name != null && this.name.toString().equals(name))) {
            if (parameters == null) {
                if (args.size() == 0) return true;
                else return false;
            }
            if (parameters.size() != args.size())
                return false;
            // Now check that the types match
            for (int n = 0; n < args.size(); n++) {
                Parameter p = (Parameter)parameters.get(n);
                if (p == null) {
                    System.out.println("Null parameter "+n+" in "+name+parameters);
                    return false;
                } else if (args.get(n) == null) {
                    System.out.println("Null arg "+n+" in "+name+args);
                    return false;
                } else if (!((Type)args.get(n)).conforms(p.type, typeenv))
                    return false;
            }
            return true;
        }
        return false;
    }

    public boolean isFinal() {
        return modifiers.modifiers.contains("FINAL");
    }

    public boolean isAbstract() {
        return modifiers.modifiers.contains("ABSTRACT");
    }

    /**
     * List of modifiers.
     * Can be null.
     */
    protected ModifierList modifiers;

    /**
     * Type of this method.
     */
    protected Type type;

    /**
     * Name of this method.
     */
    protected Identifier name;

    /**
     * List of parameters.
     * Can be null.
     */
    protected LinkedList parameters;

    /**
     * List of thrown classes.
     * Can be null.
     */
    protected LinkedList thrown_classes;

    /**
     * Body of the method.
     * Can be null.
     */
    protected StatementList body;
}
