/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.io.*;
import java.util.*;

import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;

/**
 * Base class for any abstract syntax tree node class and 
 * defines basic methods to print out constructors.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class Node {
//------------- CONSTRUCTORS
   /**
    * Creates a Node object with a blank constructor.
    */
   public Node() { 
       sourceline = last_line_parsed;
   }

   /**
    * Creates a Node object and sets the constructor.
    */
   public Node( String s ) {
       this();
       constructor = s; 
   }

    //---------- ACCESS
    /** Returns the imports in effect for this node.
     */
    public List getImports() {
        if (parent == null) {
            return new LinkedList();
        }
        return parent.getImports();
    }

    /** Returns the current class (this), if any
     */
    public ClassDefiner getThis() {
        if (parent != null) {
            /* Can't just override this in ClassDefiners, as they
             * themselves have their enclosing class as this */
            if (parent instanceof ClassDefiner) return (ClassDefiner)parent;
            return parent.getThis();
        }
        else return null;
    }
//------------- PRINTING
    // OutputStream is deprecated these days, and if 
    // we want to use anything other than standard output, 
    // we need this to be a PrintWriter.
    protected static PrintWriter outputHandle;

    // For convenience in switching back to standard output.
    protected static PrintWriter standardOutputHandle;
    static {
        standardOutputHandle = new PrintWriter(System.out);
        outputHandle = standardOutputHandle;
    }

    // For when we're doing recursive calls from within IR, we need to
    // be able to do nested setOutputToString.
    private static Stack handleStack = new Stack();

    /** 
     * Sets the Ast classes to print out to a file.
     *
     * @param filename path to a file
     */
    public static void setOutputToFile( String filename ) {
        if (outputHandle != null) handleStack.push(outputHandle);
        try {
            outputHandle = new PrintWriter(new FileOutputStream(filename));
        }
        catch (IOException e) {
            System.err.println( "Couldn't set Ast output to " 
                                + filename + ": " + e.toString() );
            setOutputToSystemOut();
        }
    }

    /**
     * Sets the Ast classes to print to standard output.
     */
    public static void setOutputToSystemOut() {
        if (outputHandle != null) handleStack.push(outputHandle);
        outputHandle = standardOutputHandle;
    }

    /**
     * Sets the Ast classes to print to the given string buffer.
     */
    public static void setOutputToString(StringWriter buffer) {
        if (outputHandle != null) handleStack.push(outputHandle);
        outputHandle = new PrintWriter(buffer);
    }

    /**
     * Tells the Ast to flush output, and also to close the stream and 
     * switch back to standard output.
     */
    public static void endOutput() {
        PrintWriter handle = outputHandle;
        if (handleStack.size() > 0) 
            outputHandle = (PrintWriter)handleStack.pop();
        try {
            handle.write("\n");
            handle.flush();
        }
        catch (Exception e) {
            System.err.println( "Couldn't flush output: " + e.toString() );
            return;
        }
        try {
            handle.close();
        }
        catch (Exception e) {
            System.err.println( "Couldn't close file: " + e.toString() );
            return;
        }
    }

    //--------------- SOURCE PRINTING
    /** Prints to stdout some Java source code that corresponds to the
     *  AST.  Comments and formatting will be gone, of course.
     */
    public void toSource() {
        print("/* UNIMPLEMENTED: "+this.getClass().getName()+" */");
    }

    /**
    * Prints out an object.
    */
   public static void print( Object s ) {
      outputHandle.print( s );
   }

   /**
    * Prints out an integer. 
    */
   public static void print( int s ) {
      outputHandle.print( s );
   }

   /**
    * Prints out an empty  list along with an comment saying what it is.
    * 
    * @param c   comment to include (no comment delimiters)
    */
   public static void printEmptyList( String c )
   {
       outputHandle.print( "new ArrayMonoList(0) /* " + c + "*/" );
   }

   /**
    * Prints out an indent according to the current indentation level.
    */
   public static void indent() {
      for( int i = 0; i < indentation_level; i++ ) {
         outputHandle.print( indent );
      }
   }

    /** 
     * Forces a linebreak, adding indentation 
     */
    public static void linebreak() {
        print("\n");
        indent();
    }

    /**
     * Prints out an opening parenthesis and increments the indentation level.
     */
    public static void oparen() {
        print( "(" );
        indentation_level++;
    }

    /**
     * Prints out a closing parenthesis and decrements the indentation level.
     */
    public static void cparen() {
        print( ")" );
        indentation_level--;
    }

    /**
     * Prints out an opening curlybrace and increments the indentation level.
     */
    public static void ocurly() {
        print( "{" );
        indentation_level++;
    }

    /**
     * Prints out a closing curlybrace and decrements the indentation level.
     */
    public static void ccurly() {
        print( "}" );
        indentation_level--;
    }

    /**
    * Prints out an open bracket, increments the indentation level, 
    * and sets up returns after commas. 
    */
   public static void obrack() {
      print( "[" );
      indentation_level++;
   }

   /**
    * Prints out a close bracket, decrements the indentation level, 
    * and turns off returns after commas. 
    */
   public static void cbrack() {
      print( "]" );
      indentation_level--;
   }
   
    /**
     * Prints out a list of nodes as source.
     * 
     * @param l    list of AST nodes to print out (may be empty)
     * @param sep  string inserted between elements
     */
    public static void printSourceList( List l, String sep ) {
        if ( l == null ) {
            return;
        }

        Iterator iter = l.iterator();
        while ( iter.hasNext() ) {
            Node n = (Node)iter.next();
            n.toSource();
            if (iter.hasNext())
                print(sep);
        }
    }

    /**
     * Prints out a list of source, including correctly outputting unknowns
     * FIXME: Doesn't really handle unknowns yet.
     * 
     * @param l    list of AST nodes to print out (may be empty)
     */
    public static void printSourceListWithUnknowns( List l ) {
        printSourceList(l);
    }
     
    /**
     * Prints out a list of nodes as source.
     * 
     * @param l    list of AST nodes to print out (may be empty)
     */
    public static void printSourceList( List l ) {
        printSourceList(l, "");
    }

    /** Prints out an argument list as source.
     */
    public static void printSourceArgumentList( List l) {
        oparen();
        printSourceList(l, ", ");
        cparen();
    }

    // ------------------ Intermediate Representation -----------------

    /** Returns an IR representation for this node */
    public IR toIR() {
        throw new Error("toIR not implemented for "+this);
    }

    /** Turns a list of nodes into a list of IRs */
    public IRList toIRList(MonoList list) {
        MonoList l = new ArrayMonoList();
        if (list != null)
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                l.add(((Node)i.next()).toIR());
            }
        return IR.make(l);
    }

    /** Turns a list of nodes into a list of string IRs */
    public IRList toIRStringList(List list) {
        MonoList l = new ArrayMonoList();
        if (list != null)
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                l.add(IR.make(i.next().toString()));
            }
        return IR.make(l);
    }

    /** Turns a list of nodes into a list of IRs */
    public IRList toIRList(List list) {
        MonoList l = new ArrayMonoList();
        if (list != null)
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                l.add(((Node)i.next()).toIR());
            }
        return IR.make(l);
    }

    protected final void setParent(Node n) {
        if (n != null) n.parent = this;
    }

    /** Sets each node in the list to have 'this' as parent 
     * @param l A list of nodes.
     */
    protected final void setParents(List l) {
        if (l == null) return;
        for (ListIterator i = l.listIterator(0); i.hasNext(); ) {
            Node n = (Node)i.next();
            n.parent = this;
        }
    }

    /** Replaces a child with another node. 
     * @param old_child old child to replace
     * @param new_child new child to replace old child with
     * @return true if the node was replaced, false otherwise.
     */ 
    public boolean replace(Node old_child, Node new_child) {
        if (old_child != new_child) {
            return replaceUC(old_child, new_child);
        }

        return false;
    }

    /** 
     * Called within 'replace'; assumes that it's okay to just 
     * go ahead and replace the kiddies.
     * @return true if the node was replaced, false otherwise.
     */
    protected boolean replaceUC(Node old_child, Node new_child) {
        return false;
    }

    /** Replaces a child in a linked list with another node. 
     * @param l linked list of nodes to replace in
     * @param old_child old child to replace
     * @param new_child new child to replace old child with
     * @return true if the node was replaced, 
     *         false otherwise.
     */ 
    protected boolean replaceInList(List l, 
            Node old_child, Node new_child) 
    {
        if (l == null) return false;
        return replaceInList(l, l.indexOf(old_child), new_child);
    }

    /** Replaces a child with a specific index in a linked list 
     *  with another node. 
     * @param l linked list of nodes to replace in (can be null)
     * @param i index of old child to replace
     * @param new_child new child to replace old child with
     * @return true if the Node was replaced,
     *         false otherwise.
     */ 
    private boolean replaceInList(List l, int index, Node new_child) 
    {
        if ((l == null) || (index <= -1)) {
            return false;
        }
        else {
            if (new_child == null) {
                // remove instead of setting
                l.remove(index);
            }
            else {
                l.set(index, new_child);
                setParent(new_child);
            }
            return true;
        }
    }

    /** Replaces a child in a linked list with another node. 
     * @param l linked list of nodes to replace in
     * @param old_child old child to replace
     * @param new_children list of new children to replace old child with
     * @return true if the node was replaced, 
     *         false otherwise.
     */ 
    protected boolean replaceInList(List l, 
                                    Node old_child, List new_children) 
    {
        if (l == null) return false;
        return replaceInList(l, l.indexOf(old_child), new_children);
    }

    /** Replaces a child with a specific index in a linked list 
     *  with a list of nodes. 
     * @param l linked list of nodes to replace in (can be null)
     * @param i index of old child to replace
     * @param new_child list of new children to replace old child with
     * @return true if the Node was replaced,
     *         false otherwise.
     */ 
    private boolean replaceInList(List l, int index, List new_children) 
    {
        if ((l == null) || (index <= -1)) {
            return false;
        }
        else {
            l.remove(index);
            if (new_children != null)
                l.addAll(index, new_children);
            return true;
        }
    }


    //---------------- INSTANCE VARIABLES
   /**
    * Name of the constructor to use.
    */
   protected String constructor;

    /**
     * The parent node, or null for the top node.
     */
    protected Node parent;

    /**
     * The source line that this node was on.
     */
    protected int sourceline;

    //----------------- CLASS VARIABLES
   /**
    * Indent string.
    */
    private final static String indent = "  ";
   /** 
    * Current indentation level.
    */
    private static int indentation_level = 0;

    /**
     * Whether to use single-element overloaded methods
     */
    protected static boolean use_singles = true;

    /**
     * This variable is set by the lexer during parsing
     */
    public static int last_line_parsed;
}
