/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;
import uiuc.Jumbo.Util.*; 

/**
 * Represents a switch statement.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class SwitchStatement extends Statement implements Labelable {
   /**
    * @param g      optional list of case groups
    * @param l      optional list of case labels
    */
   public SwitchStatement( LinkedList g, LinkedList l ) { 
      if ( g != null ) {
         case_groups = g;

         // Pick out the default case, remove it and stick it 
         // in a special place if it exists.
         Iterator iter = case_groups.iterator();
         while( iter.hasNext() ) {
            SwitchCaseBlock scb = (SwitchCaseBlock)iter.next();
            setParent(scb);
            if ( scb.isDefault() ) {
               iter.remove();
               default_block = scb;
            }
         }

      }
      else {
         // empty case_groups
         case_groups = new LinkedList();
      }

      // Add any bare labels at the end as a new case group with 
      // an empty block.
      if ( l != null ) {
         SwitchCaseBlock scb = new SwitchCaseBlock( l, 
                                                    new StatementList() );
         setParent(scb);
         // check to see if it's a default case group
         if ( scb.isDefault() ) {
             default_block = scb;
         }
         else {
            // otherwise, just add to the other case groups
            case_groups.addLast( scb );
         }
      }

      // set up an empty default case group if we haven't found one already.
      if ( default_block == null ) {
          default_block = new SwitchCaseBlock( new StatementList() );
          setParent(default_block);
      }
   }

   protected boolean replaceUC(Node old_child, Node new_child) {
       if (old_child == test) {
           setTestExpression((Expression)new_child);
           return true;
       }
       else if (old_child == default_block) {
           replaceInList(case_groups, old_child, new_child);
           default_block = (SwitchCaseBlock)new_child;
           setParent(default_block);
           return true;
       }
       else {
           return replaceInList(case_groups, old_child, new_child);
       }
   }

    public IR toIR() {
        MonoList list = new ArrayMonoList();
        for (Iterator i = case_groups.iterator(); i.hasNext(); ) {
            SwitchCaseBlock scb = (SwitchCaseBlock)i.next();
            list.add(scb.toIR());
        }
        return IR.make("switchNoLabel", new ArrayMonoList()
                       .add(test.toIR())
                       .add(IR.makeCaseList(list).flatten())
                       .add(default_block.toIR()));
    }

    public IR toIR(String label) {
        MonoList list = new ArrayMonoList();
        for (Iterator i = case_groups.iterator(); i.hasNext(); ) {
            SwitchCaseBlock scb = (SwitchCaseBlock)i.next();
            list.add(scb.toIR());
        }
        return IR.make("switchLabel", new ArrayMonoList()
                       .add(IR.make(label))
                       .add(test.toIR())
                       .add(IR.makeCaseList(list).flatten())
                       .add(default_block.toIR()));
    }

    public void toSource() {
        print("switch (");
        test.toSource();
        print(") ");
        ocurly();
        linebreak();
        for (Iterator i = case_groups.iterator(); i.hasNext(); ) {
            SwitchCaseBlock scb = (SwitchCaseBlock)i.next();
            scb.toSource();
        }
        if (default_block != null)
            default_block.toSource();
        ccurly();
        linebreak();
    }

    public boolean needsSemicolon() { return false; }

   /**
    * Sets the test expression.
    * 
    * @param e  new test expression
    */
   public void setTestExpression( Expression e ) {
      test = e;
      setParent(test);
   }

    public String getLabel() {
        if (parent instanceof LabeledStatement)
            return ((LabeledStatement)parent).getLabel();
        return "";
    }

   /**
    * Test expression.
    */
   protected Expression test;
   /**
    * List of case groups.
    */
   protected LinkedList case_groups;
   /**
    * The default case.
    */
   protected SwitchCaseBlock default_block;
}
