/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

/**
 * Represents a type. 
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

import java.util.*;

public class Type extends Node {
    /** A type that is not determined yet */
    public static final byte UNKNOWN    = -1;
    /** A type for NULL */
    public static final byte NULL    = 0;
    /*public static final byte ARRAY   = 1;*/
    public static final byte OBJECT  = 2;
    public static final byte VOID    = 3;
    public static final byte BOOLEAN = 4;
    public static final byte CHAR    = 5;
    public static final byte FLOAT   = 6;
    public static final byte DOUBLE  = 7;
    public static final byte BYTE    = 8;
    public static final byte SHORT   = 9;
    public static final byte INT     = 10;
    public static final byte LONG    = 11;

    private int kind;
    
    /** The name of this type, if it is an object type */
    private String name;

    /** The dimensions of this type (0 = not an array) */
    private int dimensions;

    /**
     * Creates a completely blank primitive type for 
     * filling in later.
     */
    public Type() {
        this( "", 0 );
    }

    /**
     * Creates a type with no dimensions.
     *
     * @param tname   name of the type
     */
    public Type( Object tname ) { 
        this( tname, 0 );
    }

    public Type( Object tname, int dims ) {
        setName(tname);
        dimensions = dims;
    }

    /** Create a type with specific kind, name and dimensions. */
    protected Type( int kind, String tname, int dims ) {
        this.kind = kind;
        this.name = tname;
        this.dimensions = dims;
    }

    public Type copyType() {
        Type t = new Type(kind, name, dimensions);
        return t;
    }

    /**
     * Increments the number of dimension by 1.
     */
    public void incrementDimensions() {
        dimensions += 1;
    }

    /**
     * Sets the number of dimensions.
     *
     * @param dims  Number of dimensions to set this type to
     */
    public void setDimensions( int dims ) {
        dimensions = dims;
    }

    /**
     * Sets the name of the type.
     *
     * @param tname  New name of this type
     */
    public void setName( Object tname ) {
        if (tname == null || tname.equals("")) {
            kind = UNKNOWN;
        } else {
            name = null;
            if (tname.equals("void")) kind = VOID;
            else if (tname.equals("boolean")) kind = BOOLEAN;
            else if (tname.equals("byte")) kind = BYTE;
            else if (tname.equals("char")) kind = CHAR;
            else if (tname.equals("short")) kind = SHORT;
            else if (tname.equals("int")) kind = INT;
            else if (tname.equals("long")) kind = LONG;
            else if (tname.equals("float")) kind = FLOAT;
            else if (tname.equals("double")) kind = DOUBLE;
            else {
                kind = OBJECT;
                name = tname.toString();
            }
        }
    }

    /**
     * Creates a Jaemus.Type object for this type
     */
    public uiuc.Jumbo.Jaemus.Type toType() {
        uiuc.Jumbo.Jaemus.Type t;
        switch (kind) {
        case NULL: t = uiuc.Jumbo.Jaemus.Type.null_type; break;
        case OBJECT: t = new uiuc.Jumbo.Jaemus.Type(name); break;
        case VOID: t = uiuc.Jumbo.Jaemus.Type.void_type; break;
        case BOOLEAN: t = uiuc.Jumbo.Jaemus.Type.boolean_type; break;
        case CHAR: t = uiuc.Jumbo.Jaemus.Type.char_type; break;
        case FLOAT: t = uiuc.Jumbo.Jaemus.Type.float_type; break;
        case DOUBLE: t = uiuc.Jumbo.Jaemus.Type.double_type; break;
        case BYTE: t = uiuc.Jumbo.Jaemus.Type.byte_type; break;
        case SHORT: t = uiuc.Jumbo.Jaemus.Type.short_type; break;
        case INT: t = uiuc.Jumbo.Jaemus.Type.int_type; break;
        case LONG: t = uiuc.Jumbo.Jaemus.Type.long_type; break;
        default: throw new Error("Unknown kind "+kind);
        }
        for (int i = 0; i < dimensions; i++) {
            t = new uiuc.Jumbo.Jaemus.Type(t);
        }
        return t;
    }

    public IR toIR() {
        return IR.make(toType());
    }

    private String[] typenames = {
        "null",
        "array",
        "object",
        "void",
        "boolean",
        "char",
        "float",
        "double",
        "byte",
        "short",
        "int",
        "long"
    };

    /** 
     * Outputs this type as source
     */
    public void toSource() {
        print(toString());
    }

    public String toString() {
        String result;
        if (isPrimitive() || isVoid() || isNull())
            result = typenames[kind];
        else
            result = name;
        for (int i = 0; i < dimensions; i++) {
            result += "[]";
        }
        return result;
    }

    /** Return a Class representing this type.
     * This will only work for types representing existing classes
     * and primitive types.
     */
    public Class toClass() {
        try {
            if (isArray()) {
                String dims = "";
                for (int i = 0; i < dimensions; i++) {
                    dims += "[";
                }
                switch (kind) {
                case OBJECT:
                    try {
                        return Class.forName(dims+"L"+name+";");
                    } catch (ClassNotFoundException e) {
                        return null;
                    }
                case VOID: return Class.forName(dims+"V");
                case BOOLEAN: return Class.forName(dims+"Z");
                case CHAR: return Class.forName(dims+"C");
                case FLOAT: return Class.forName(dims+"F");
                case DOUBLE: return Class.forName(dims+"D");
                case BYTE: return Class.forName(dims+"B");
                case SHORT: return Class.forName(dims+"S");
                case INT: return Class.forName(dims+"I");
                case LONG: return Class.forName(dims+"J");
                default: throw new Error("Unclassable array kind for "+this);
                }
            }
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find class for array: "+e);
        }
        switch (kind) {
        case OBJECT: 
            try {
                return Class.forName(name);
               
            } catch (ClassNotFoundException e) {
                return null;
            }
        case VOID: return Void.TYPE;
        case BOOLEAN: return Boolean.TYPE;
        case CHAR: return Character.TYPE;
        case FLOAT: return Float.TYPE;
        case DOUBLE: return Double.TYPE;
        case BYTE: return Byte.TYPE;
        case SHORT: return Short.TYPE;
        case INT: return Integer.TYPE;
        case LONG: return Long.TYPE;
        default: throw new Error("Unclassable kind for "+this);
        }
        }

    /**
     * Returns whether this type is an array type or not.
     *
     * @return True if this type is an array type, false otherwise.
     */
    public boolean isArray() {
       return (dimensions > 0);
    }


    /**
     * Returns whether this type is a primitive type or not.
     * This also returns true for an array of primitive types.
     *
     * @return True if this type is a primitive type, false otherwise.
     */
    public boolean isPrimitive() {
        return kind > VOID;
    }

    /**
     * Returns true if this type is a reference type (object, array or null).
     */
    public boolean isReference() {
        return kind <= OBJECT || dimensions > 0;
    }
    
    /**
     * Returns whether this type is the 'void' type.
     *
     * @return True if this type is 'void', false otherwise.
     */
    public boolean isVoid() {
        return kind == VOID;
    }

    /** Returns true if this type is the type of 'null'.
     */
    public boolean isNull() {
        return kind == NULL;
    }

    /**
     * Returns a new type representing the type of the contents of this array.
     */
    public Type getElementType() {
        if (dimensions == 0) throw new Error("Taking element type of non-array");
        return new Type(kind, name, dimensions-1);
    }

    /** Returns the name in this type, if non-primitive.
     * For an array type, returns the name of the array element.
     */
    public String getName() {
        if (isPrimitive()) throw new Error("Taking class name of primitive type");
        return name;
    }

    public int getKind() {
        return kind;
    }

    /** May this primitive type be widened to the given primitive kind? */
    public boolean mayWidenPrimitive(int kind) {
        if (this.kind == kind) return true;
        switch (this.kind) {
        case BOOLEAN:
            return false; // Only boolean conforms to boolean
        case BYTE:
            return (kind == SHORT ||
                    kind == INT ||
                    kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case SHORT: 
        case CHAR:
            return (kind == INT ||
                    kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case INT:
            return (kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case LONG:
            return (kind == FLOAT ||
                    kind == DOUBLE);
        case FLOAT:
            return (kind == DOUBLE);
        case DOUBLE:
            return false; // Only double conforms to double
        default:
            return false;
        }
    }
    
    /** Check that this type conforms (methodwise) to the given type.
     * @param t A Type.
     * @param superfinder An implementation of SuperFinder.
     * @return True if t is the same type as this, or a supertype.
     */
    public boolean conforms(Type t, TypeEnv typeenv) {
        if (isArray()) {
            if (t.dimensions > dimensions) return false;
            if (t.dimensions < dimensions) {
                // t's elements must be Object
                if (t.name.equals("Object") ||
                    t.name.equals("java.lang.Object"))
                    return true;
                return false;
            }
            // Otherwise, we must check the objects anyway
        }
        // If the array dims match, check the elements
        if (isPrimitive()) {
            // Check for casting of primitive types
            return mayWidenPrimitive(t.kind);
        }
        // Objects must check for subclassing, even inside arrays
        Type othert = typeenv.getFQCN(t);
        Type thist = typeenv.getFQCN(this);
        if (othert == null) return false;
        if (thist == null) throw new Error("Null thist for "+this);
        if (thist.getName() == null) throw new Error("Null getName for "+thist);
        if (thist.getName().equals(othert.getName())) return true;
        List supers = typeenv.getSupers(this);
        for (Iterator i = supers.iterator(); i.hasNext(); ) {
            Type supert = (Type)i.next();
            if (supert.conforms(othert, typeenv)) return true;
        }
        return false;
    }
}


/** Interface for functions for finding environment information on Types.*/
interface TypeEnv {
    /** Returns the list of superclass types (including interfaces) */
    List getSupers(Type t);
    /** Returns the fully qualified class name for the given type */
    Type getFQCN(Type t);
}
