/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;
import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;

/**
 * Represents a variable.  
 *
 * Happens to be packaged with a type as well as an identifier for 
 * variable declarations like <code>int a, b[][][][]</code>.
 *
 * The type effectively always has to exist, even if it's blank, 
 * because of funky variable declarations like the above.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class Variable extends Node {
    /**
     * Creates a variable with a blank type.
     *
     * @param i      Name of this variable
     */
    public Variable( Identifier i ) {
        // Just tracing!
        this( i, new Type(i.toString()) );
    }

    public Variable( String i ) {
        // Just tracing!
        this( new Identifier(i), new Type(i) );
    }

    /**
     * @param i      Name of this variable
     * @param t      Type of this variable
     */
    public Variable( Identifier i, Type t) { 
        setName(i);
        setType(t);
    }

    /** Make this variable be one dimension bigger.
     * I.e. int -> int[], String[] -> String[][]
     * Necessary because you can say int[] x[] for int[][] x:(
     */
    public void makeArray() {
        dimensions++;
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
        if (old_child == name) {
            setName((Identifier)new_child);
            return true;
        }
        else if (old_child == type) {
            setType((Type)new_child);
            return true;
        }
        else {
            return false;
        }
    }

    protected void setName(Identifier i) {
        name = i;
        setParent(name);
    }

    protected void setType(Type t) {
        type = t;
        for (int i = 0; i < dimensions; i++) 
            t.incrementDimensions();
        setParent(type);
    }

    /**
     * Prints out the last part of the source needed to 
     * make a variable declaration (name followed by type).
     * Assumes that it is called at the end of a parent declaration.
     */
    public void toDeclarationSource() {
        name.toSource();
    }

    public IR toIR() {
        return IR.make("getQualifiedName", new ArrayMonoList()
                       .add(name.toIRString()));
    }

    /** Returns a string description of this variable.
     */
    public String toString() {
        return type+" "+name;
    }

    /**
     * Name of this variable.
     */
    protected Identifier name;
    /**
     * Type of this variable.
     * Public for easy modification access.
     */
    public Type type;

    /**
     * Number of dimensions inherent to this variable
     */
    private int dimensions;
}
